# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.templates.doc_templates`` contains the definition of
the document template classes used in the PDF generation with ReportLab
`platypus`.

These classes are derived from platypus standard classes but are specific to
PyBill software. The document template is used to create new PDF
documents for the PyBill accounting documents. It adds some behaviour to its
base class in the flow content management, specifically the ability to compute
the carry forward from a :class:`~pybill.lib.pdfwriters.utils.TotalizableTable`
inserted in the flow content.

.. autoclass:: AccDocDocumentTemplate
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from reportlab.platypus import BaseDocTemplate

from pybill.lib.pdfwriters.utils import TotalizableTable
from pybill.lib.pdfwriters.templates.page_templates import (AccDocPageTemplate,
                                                       AccDocFirstPageTemplate)


class AccDocDocumentTemplate(BaseDocTemplate):
    """
    The accounting document template is a class explaining how to create an
    accouting PDF document.

    It uses the
    :class:`~pybill.lib.pdfwriters.templates.page_templates.AccDocPageTemplate`
    and the
    :class:`~pybill.lib.pdfwriters.templates.page_templates.AccDocFirstPageTemplate`
    page templates and can compute the carry forward that has to be brought
    forward from one page to another.

    The computation of the carry forward is only possible if the main table
    of the document is a :class:`~pybill.lib.pdfwriters.utils.TotalizableTable`
    and if, in this table, the total of each row is put in a
    :class:`~pybill.lib.pdfwriters.utils.NumberParagraph`.
    
    The carry forward will not appear if a
    :class:`~pybill.lib.pdfwriters.utils.TotalizableTable` is not used.

    .. attribute:: carry_forward

       Attribute containing the carry forward that has to be brought forward
       from a finished page to the next page.

       type: :class:`float`

    .. attribute:: _carry_forward_computed

       Attribute containing a flag set to ``True`` if the carry forward has once
       been computed.

       type: :class:`bool`
    
    .. attribute:: display_carry_forward

       Attribute containing a flag set to ``True`` if the carry forward has to
       be displayed at end of the page (and the beginning of the next page).
                                 
       type: :class:`bool`
    
    .. attribute:: all_flowables_handled

       Attribute containing a flag set to ``True`` if all the flowables have
       been handled.

       The carry forward must not be displayed on the last page. The only way I
       found to know if is the last page, is to watch the build progress and set
       a flag when I see the last flowable has been handled. Anyway, this works
       properly because the last flowable of the document is in a
       :class:`~reportlab.platypus.KeepTogether` and is never splitted over
       pages.
        
       type: :class:`bool`

    .. attribute:: _estim_num_of_flowables

       Attribute containing the estimated number of flowables to handle (see
       :attr:`all_flowables_handled` for more details).
       
       type: :class:`int`

    .. automethod:: __init__
    """
    def __init__(self, filename, acc_doc):
        """
        Initializes an ``AccDocDocumentTemplate`` object.

        This method initializes the correct page templates to be used when
        generating the document, sets the PDF document metadata, and initializes
        the attributes.
        
        :parameter filename: name of the PDF output file
        :type filename: unicode

        :parameter acc_doc: Accounting entity containing the actual document
                            to be rendered in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`
        """
        # Initializes the document template with the correct page templates.
        accdoc_templates = [AccDocFirstPageTemplate(acc_doc,
                                                    template_id="First"),
                            AccDocPageTemplate(acc_doc,
                                               template_id="Next")]
        BaseDocTemplate.__init__(self, filename,
                                 pageTemplates=accdoc_templates,
                                 allowSplitting=1)
        # Sets the PDF metadata (initialized by base class).
        self.title = u" ".join(acc_doc.doc_ref.split())
        self.author = acc_doc.cfg_data.company_address.name.strip()
        # Specific attributes.
        self.carry_forward = 0.0
        self._carry_forward_computed = False
        self.display_carry_forward = False
        self.all_flowables_handled = False
        self._estim_num_of_flowables = 0
        # Registers the progress function that will set the
        # ``all_flowables_handled`` flag.
        self._onProgress = self.watch_flowable_handling

    def handle_pageBegin(self):
        """
        Overrides base method to add a change of page template after the
        first page and to unset the :attr:`display_carry_forward` flag.
        """
        self._handle_pageBegin()
        if self.page == 1:
            self._handle_nextPageTemplate("Next")
        self.display_carry_forward = False

    def handle_pageEnd(self):
        """
        Overrides base method to set the :attr:`display_carry_forward` flag
        when we are ending a page (that is not the last page) and when the
        main table has already been displayed (or is being displayed).
        """
        if not self.all_flowables_handled:
            # If the carry forward hasn't been computed (``TotalizableTable``
            # not yet displayed on the pages), don't display it.
            self.display_carry_forward = self._carry_forward_computed
        self._handle_pageEnd()

    def afterFlowable(self, flow):
        """
        Overrides base method to compute the carry forward if the flowable
        is a :class:`~pybill.lib.pdfwriters.utils.TotalizableTable`.

        The only :class:`~pybill.lib.pdfwriters.utils.TotalizableTable` in the
        flowables list is the main table that will be used to compute the
        carry forward.
        
        :parameter flow: Flowable that has just been drawn on the page.
        :type flow: :class:`reportlab.platypus.Flowable`
        """
        if isinstance(flow, TotalizableTable):
            self._carry_forward_computed = True
            self.carry_forward += flow.compute_total()

    def watch_flowable_handling(self, ev_type, value):
        """
        Progress function that is called by the document template as it builds
        the PDF document.

        This function has been registered by the :meth`__init__` method in the
        hook called during document process (cf. :attr:`_onProgress` attribute
        of the base class). It sets the :attr:`all_flowables_handled` flag.

        :parameter ev_type: Type of the event. Possible types are:
        
                            - ``PASS``: number of the pass starting
                              now. Numerous passes are done when working with
                              multi-build for indexing flowables.
                              
                            - ``STARTED``: starting build.
                            
                            - ``SIZE_EST``: estimated number of flowables to be
                              handled (some flowables may be added when one
                              flowable is splitted over pages).
                              
                            - ``PROGRESS``: number of the flowable that has just
                              been handled (if a flowable is splitted over
                              pages, its number will appear various times in
                              PROGRESS).
                                      
                            - ``PAGE``: number of the page that has just been
                              finished.
                                  
                            - ``FINISHED``: ending build.
        :type ev_type: :class:`str`

        :parameter value: Number related to the event.
        :type value: :class:`int`
        """
        if ev_type == "SIZE_EST":
            self._estim_num_of_flowables = value
        # The last flowable in an accounting document (the total table) is in a
        # ``KeepTogether``, thus we know it will never be splitted over two
        # pages. Therefore when we see the last flowable in `PROGRESS` we know
        # there is no more flowable to handle (clever isn't it).
        if ev_type == "PROGRESS":
            self.all_flowables_handled = (self._estim_num_of_flowables == value)



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.templates`` sub-package contains the definition of the
template classes (page templates and document template) used in the PDF
generation with ReportLab `platypus`.

The page templates describe the structure of the PDF pages and where the flow
content can be written. The document template describes a PDF document, mainly
explaining which page templates will be used. Both templates are used by the
platypus machinery to display, into the PDF documents, the flowables generated
by the :mod:`pybill.lib.pdfwriters.flowables` flow generators.

The sub-package is divided in several modules, each defining one type of
template.

``page_templates`` module
=========================

.. automodule:: pybill.lib.pdfwriters.templates.page_templates

``doc_templates`` module
========================

.. automodule:: pybill.lib.pdfwriters.templates.doc_templates
"""
__docformat__ = "restructuredtext en"


# Imports the classes that are exported from this package
from pybill.lib.pdfwriters.templates.page_templates import (AccDocPageTemplate,
                                                       AccDocFirstPageTemplate)
from pybill.lib.pdfwriters.templates.doc_templates import AccDocDocumentTemplate



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from lxml import etree
from StringIO import StringIO
from tempfile import mkdtemp
from os import remove, listdir, rmdir
from os import path as osp

from pybill.lib.entities.addresses import PersonAddress, OrganisationAddress
from pybill.lib.xmlwriters.utils import (build_address_xml_element,
                                         open_xml_outfile)


class AccDocFormatAddressWriterTest(TestCase):
    def test_complete_address_writing(self):
        """
        Tests the writing of an address that has all of its fields filled.
        """
        # Address building
        addr = PersonAddress()
        addr.honorific = u"A1"
        addr.firstname = u"A2"
        addr.other_name = u"A3"
        addr.surname = u"A4"
        addr.lineage = u"A5"
        addr.streets = [u"A6" , u"A7"]
        addr.post_box = u"A8"
        addr.post_code = u"A9"
        addr.city = u"A10"
        addr.state = u"A11"
        addr.country = u"A12"
        addr.phone = u"A13"
        addr.fax = u"A14"
        addr.web = u"A15"
        addr.email = u"A16"
        # affiliation
        addr.organisation = OrganisationAddress()
        addr.organisation.name = u"A17"
        addr.organisation.divisions = [u"A18", u"A19"]
        addr.organisation.job_titles = [u"A20", u"A21"]
        addr.organisation.streets = [u"A22", u"A23"]
        addr.organisation.post_box = u"A24"
        addr.organisation.post_code = u"A25"
        addr.organisation.city = u"A26"
        addr.organisation.state = u"A27"
        addr.organisation.country = u"A28"
        addr.organisation.phone = u"A29"
        addr.organisation.fax = u"A30"
        addr.organisation.web = u"A31"
        addr.organisation.email = u"A32"
        # Expected result (XML)
        expected_xml = u"""<address>
  <honorific>A1</honorific>
  <firstname>A2</firstname>
  <othername>A3</othername>
  <surname>A4</surname>
  <lineage>A5</lineage>
  <street>A6</street>
  <street>A7</street>
  <pob>A8</pob>
  <postcode>A9</postcode>
  <city>A10</city>
  <state>A11</state>
  <country>A12</country>
  <phone>A13</phone>
  <fax>A14</fax>
  <web>A15</web>
  <email>A16</email>
  <affiliation>
    <orgname>A17</orgname>
    <orgdiv>A18</orgdiv>
    <orgdiv>A19</orgdiv>
    <jobtitle>A20</jobtitle>
    <jobtitle>A21</jobtitle>
    <address>
      <street>A22</street>
      <street>A23</street>
      <pob>A24</pob>
      <postcode>A25</postcode>
      <city>A26</city>
      <state>A27</state>
      <country>A28</country>
      <phone>A29</phone>
      <fax>A30</fax>
      <web>A31</web>
      <email>A32</email>
    </address>
  </affiliation>
</address>
"""
        xml_elt = build_address_xml_element(addr)
        self.assertMultiLineEqual(etree.tostring(xml_elt, pretty_print=True),
                              expected_xml)

    def test_no_affiliation_address_writing(self):
        """
        Tests the writing of an address with no organisation address embedded
        and thus no affiliation.
        """
        # Address building
        addr = PersonAddress()
        addr.honorific = u"A1"
        addr.firstname = u"A2"
        addr.other_name = u"A3"
        addr.surname = u"A4"
        addr.lineage = u"A5"
        addr.streets = [u"A6" , u"A7"]
        addr.post_box = u"A8"
        addr.post_code = u"A9"
        addr.city = u"A10"
        addr.state = u"A11"
        addr.country = u"A12"
        addr.phone = u"A13"
        addr.fax = u"A14"
        addr.web = u"A15"
        addr.email = u"A16"
        # Expected result (XML)
        expected_xml = u"""<address>
  <honorific>A1</honorific>
  <firstname>A2</firstname>
  <othername>A3</othername>
  <surname>A4</surname>
  <lineage>A5</lineage>
  <street>A6</street>
  <street>A7</street>
  <pob>A8</pob>
  <postcode>A9</postcode>
  <city>A10</city>
  <state>A11</state>
  <country>A12</country>
  <phone>A13</phone>
  <fax>A14</fax>
  <web>A15</web>
  <email>A16</email>
</address>
"""
        xml_elt = build_address_xml_element(addr)
        self.assertMultiLineEqual(etree.tostring(xml_elt, pretty_print=True),
                              expected_xml)

    def test_regular_address_writing(self):
        """
        Tests the writing of a common address (with few fields).
        """
        # Address building
        addr = PersonAddress()
        addr.honorific = u"A1"
        addr.firstname = u"A2"
        addr.surname = u"A3"
        addr.phone = u"A4"
        addr.email = u"A5"
        # affiliation
        addr.organisation = OrganisationAddress()
        addr.organisation.name = u"A6"
        addr.organisation.job_titles = [u"A7"]
        addr.organisation.streets = [u"A8"]
        addr.organisation.post_code = u"A9"
        addr.organisation.city = u"A10"
        # Expected result (XML)
        expected_xml = """<address>
  <honorific>A1</honorific>
  <firstname>A2</firstname>
  <surname>A3</surname>
  <phone>A4</phone>
  <email>A5</email>
  <affiliation>
    <orgname>A6</orgname>
    <jobtitle>A7</jobtitle>
    <address>
      <street>A8</street>
      <postcode>A9</postcode>
      <city>A10</city>
    </address>
  </affiliation>
</address>
"""
        xml_elt = build_address_xml_element(addr)
        self.assertMultiLineEqual(etree.tostring(xml_elt, pretty_print=True),
                              expected_xml)

    def test_complete_address_writing_with_namespace(self):
        """
        Tests the writing of an address that has all of its fields filled and
        that must be written in a given XML namespace.
        """
        # Address building
        addr = PersonAddress()
        addr.honorific = u"A1"
        addr.firstname = u"A2"
        addr.other_name = u"A3"
        addr.surname = u"A4"
        addr.lineage = u"A5"
        addr.streets = [u"A6" , u"A7"]
        addr.post_box = u"A8"
        addr.post_code = u"A9"
        addr.city = u"A10"
        addr.state = u"A11"
        addr.country = u"A12"
        addr.phone = u"A13"
        addr.fax = u"A14"
        addr.web = u"A15"
        addr.email = u"A16"
        # affiliation
        addr.organisation = OrganisationAddress()
        addr.organisation.name = u"A17"
        addr.organisation.divisions = [u"A18", u"A19"]
        addr.organisation.job_titles = [u"A20", u"A21"]
        addr.organisation.streets = [u"A22", u"A23"]
        addr.organisation.post_box = u"A24"
        addr.organisation.post_code = u"A25"
        addr.organisation.city = u"A26"
        addr.organisation.state = u"A27"
        addr.organisation.country = u"A28"
        addr.organisation.phone = u"A29"
        addr.organisation.fax = u"A30"
        addr.organisation.web = u"A31"
        addr.organisation.email = u"A32"
        # Expected result (XML)
        expected_xml = u"""<ns0:address xmlns:ns0="http://www.logilab.org/TEST">
  <ns0:honorific>A1</ns0:honorific>
  <ns0:firstname>A2</ns0:firstname>
  <ns0:othername>A3</ns0:othername>
  <ns0:surname>A4</ns0:surname>
  <ns0:lineage>A5</ns0:lineage>
  <ns0:street>A6</ns0:street>
  <ns0:street>A7</ns0:street>
  <ns0:pob>A8</ns0:pob>
  <ns0:postcode>A9</ns0:postcode>
  <ns0:city>A10</ns0:city>
  <ns0:state>A11</ns0:state>
  <ns0:country>A12</ns0:country>
  <ns0:phone>A13</ns0:phone>
  <ns0:fax>A14</ns0:fax>
  <ns0:web>A15</ns0:web>
  <ns0:email>A16</ns0:email>
  <ns0:affiliation>
    <ns0:orgname>A17</ns0:orgname>
    <ns0:orgdiv>A18</ns0:orgdiv>
    <ns0:orgdiv>A19</ns0:orgdiv>
    <ns0:jobtitle>A20</ns0:jobtitle>
    <ns0:jobtitle>A21</ns0:jobtitle>
    <ns0:address>
      <ns0:street>A22</ns0:street>
      <ns0:street>A23</ns0:street>
      <ns0:pob>A24</ns0:pob>
      <ns0:postcode>A25</ns0:postcode>
      <ns0:city>A26</ns0:city>
      <ns0:state>A27</ns0:state>
      <ns0:country>A28</ns0:country>
      <ns0:phone>A29</ns0:phone>
      <ns0:fax>A30</ns0:fax>
      <ns0:web>A31</ns0:web>
      <ns0:email>A32</ns0:email>
    </ns0:address>
  </ns0:affiliation>
</ns0:address>
"""
        xml_elt = build_address_xml_element(addr,
                                          xml_ns="http://www.logilab.org/TEST")
        self.assertMultiLineEqual(etree.tostring(xml_elt, pretty_print=True),
                              expected_xml)

class XMLOpenTest(TestCase):
    """
    Tests the function that opens an XML file and writes the XML declaration.
    """
    def test_opening_from_filename(self):
        # Creates a temporary directory
        temp_dir = mkdtemp()
        # Runs the function
        filename = osp.join(temp_dir, "test.xml")
        out = open_xml_outfile(filename)
        out.close()
        # Tests the result
        res_file = open(filename)
        result = res_file.read()
        self.assertMultiLineEqual(result,
                              '<?xml version="1.0" encoding="UTF-8"?>\n\n')
        res_file.close()
        # Removes the temporary directory
        for fname in listdir(temp_dir):
            remove(osp.join(temp_dir, fname))
        rmdir(temp_dir)

    def test_opening_from_file(self):
        # Creates an opened file
        opened_file = StringIO()
        # Runs the function
        out = open_xml_outfile(opened_file)
        # Tests the result
        self.assertEqual(out, opened_file)
        self.assertMultiLineEqual(out.getvalue(),
                              '<?xml version="1.0" encoding="UTF-8"?>\n\n')

# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()




# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlwriters`` package contains the classes and functions that
PyBill accounting entities in XML files.

Currently, PyBill just knows how to dump the entities in the
`PyBill Document 1.0` format. There is no support for writing the entities in
the old formats.

The :func:`write_accounting_doc` is the public function that should be used from
the outside of this package. The :const:`ENCODING` constant is also public and
contains the default encoding used for all the XML files written by PyBill. They
are both imported and available in the namespace of this package.
"""
__docformat__ = "restructuredtext en"


# Imports the constants and functions that must be used from the outside.
from pybill.lib.xmlwriters.utils import ENCODING
from pybill.lib.xmlwriters.accdoc import write_accounting_doc



# -*- coding: utf-8 -*-
"""
``pybill.lib.entities.accounting_docs.downpayment`` module defines the entity
representing a downpayment.

.. autoclass:: Downpayment
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib import ACCURACY
from pybill.lib.entities.accounting_docs.abstract import GenericAccountingDoc


class Downpayment(GenericAccountingDoc):
    """
    Entity class representing a downpayment

    .. attribute:: percent

       Attribute defining the percent of the total that must be paid as a
       downpayment. Hence, the downpayment represents a percent of the actual
       bill. By default, the percent is ``30.0``\ .

       type: :class:`float`

    .. attribute:: payment_terms

       Attribute defiing the description of the payment terms. This description
       will be displayed after the bank data that is read from the configuration
       file.
                         
       type: :class:`unicode`

    .. automethod:: __init__   
    """
    def __init__(self, downpay_id):
        """
        Initializes an empty object.

        :parameter downpay_id: Identifier of this accounting document.
        :type downpay_id: :class:`unicode`
        """
        GenericAccountingDoc.__init__(self)
        
        self.percent = 30.0
        self.payment_terms = u""
        # Filling the object with available data
        self.id = downpay_id

    def get_downpayment(self):
        """
        Computes the downpayment.

        The downpayment is the including taxes total multiplied by the
        downpayment percent.
        
        :returns: The amount of the downpayment.
        :rtype: :class:`float`
        """
        return round(self.get_including_taxes_total()
                     * round(self.percent, ACCURACY) / 100., ACCURACY)

    def get_to_be_paid_holdback_free(self):
        """
        Computes the total without holdback that must be paid.

        This total depends, of course, on the downpayment but also on the
        holdbacks.
        
        :returns: The total without holdback of the downpayment that must be
                  paid.
        :rtype: :class:`float`
        """
        total = self.get_including_taxes_total()
        total_dwp = self.get_downpayment()
        total_hbk = self.get_holdback_amount() + self.get_holdback_vat_amount()
        if total_dwp <= (total - total_hbk):
            return round(total_dwp, ACCURACY)
        else:
            return round(total - total_hbk, ACCURACY)

    def get_to_be_paid_holdback(self):
        """
        Computes the total of holdbacks that must be paid.

        This total depends, of course, on the downpayment but also on the
        holdbacks.
        
        :returns: The total of holdbacks of the downpayment that must be paid.
        :rtype: :class:`float`
        """
        total = self.get_including_taxes_total()
        total_dwp = self.get_downpayment()
        total_hbk = self.get_holdback_amount() + self.get_holdback_vat_amount()
        if total_dwp <= (total - total_hbk):
            return 0.0
        else:
            return round(total_dwp - (total - total_hbk), ACCURACY)

    def get_to_be_paid_vat(self):
        """
        Computes the total of VAT amounts that must be paid. This amount depends
        on the VAT amounts but also, of course, on the downpayment percent.

        :returns: The total of VAT amounts that must be paid.

        :rtype: :class:`float`
        """
        vat_amount = self.get_including_taxes_total() - self.get_price_total()
        return round(float(vat_amount * self.percent) / 100., ACCURACY)



# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlwriters.accdoc_format_1_0`` module defines the function that
can write the accounting entities in an XML element in `PyBill Document 1.0`
format.

The function can write any accounting entities such as bills, claim-forms,
debits, downpayments, or pro-formas.

.. autofunction:: write_accdoc_1_0_xml
"""
__docformat__ = "restructuredtext en"

from lxml import etree

from pybill.lib import PBD_1_0, PBD_XMLNS
from pybill.lib.entities.accounting_docs import (Bill, ClaimForm, Debit,
                                                 Downpayment, ProForma)
from pybill.lib.xmlwriters.utils import build_address_xml_element

ACCDOC_TYPE = {Bill: u"bill",
               ClaimForm: u"claim-form",
               Debit: u"debit",
               Downpayment: u"downpayment",
               ProForma: u"pro-forma"}


def write_accdoc_1_0_xml(acc_doc):
    """
    Writes an accounting entity into an XML element in `PyBill Document 1.0`
    format.

    The XML element is built thanks to `lxml` library.
    
    :parameter acc_doc: PyBill entity describing the accounting document to
                        be turned in XML. Can be a ``Bill``, a ``ClaimForm``,
                        a ``Debit``, a ``Downpayment`` or a ``ProForma``
                        object.
    :type acc_doc: :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

    :returns: an XML element describing the accounting document in
              `PyBill Document 1.0` format.
    :rtype: :class:`lxml.etree.Element`
    """
    root_elt = etree.Element(u"{%s}accounting-document" % PBD_XMLNS,
                             nsmap={"pbd": PBD_XMLNS})
    root_elt.set(u"{%s}format-version" % PBD_XMLNS, PBD_1_0)
    root_elt.set(u"{%s}type" % PBD_XMLNS, ACCDOC_TYPE[acc_doc.__class__])
    # Downpayment percent (only for downpayments)
    if acc_doc.__class__ is Downpayment:
        dwp_elt = etree.SubElement(root_elt,
                                   u"{%s}downpayment-percent"  % PBD_XMLNS)
        dwp_elt.text = unicode(acc_doc.percent)
    # Validity date (only for pro-formas)
    if acc_doc.__class__ is ProForma:
        vld_elt = etree.SubElement(root_elt,
                                   u"{%s}validity-date" % PBD_XMLNS)
        vld_elt.text = acc_doc.validity_date
    # Metadata
    mtd_elt = etree.SubElement(root_elt, u"{%s}metadata" % PBD_XMLNS)
    if acc_doc.__class__ in [Bill, ClaimForm, Debit, Downpayment]:
        id_elt = etree.SubElement(mtd_elt, u"{%s}id" % PBD_XMLNS)
        id_elt.text = acc_doc.id
    drf_elt = etree.SubElement(mtd_elt, u"{%s}doc-ref" % PBD_XMLNS)
    drf_elt.text = acc_doc.doc_ref
    plc_elt = etree.SubElement(mtd_elt, u"{%s}place" % PBD_XMLNS)
    plc_elt.text = acc_doc.place
    dte_elt = etree.SubElement(mtd_elt, u"{%s}date" % PBD_XMLNS)
    dte_elt.text = acc_doc.date
    if acc_doc.date_num is not None:
        dte_elt.set(u"{%s}num" % PBD_XMLNS,
                    unicode(acc_doc.date_num.strftime("%Y-%m-%d")))
    other_kw = acc_doc.other_infos.keys()
    other_kw.sort() # Just to ease the tests
    for kw in other_kw:
        inf_elt = etree.SubElement(mtd_elt, u"{%s}info"  % PBD_XMLNS)
        inf_elt.set("{%s}name" % PBD_XMLNS, kw)
        inf_elt.text = acc_doc.other_infos[kw]
    # Sender address
    if acc_doc.sender is not None:
        sdr_elt = build_address_xml_element(acc_doc.sender, xml_ns= PBD_XMLNS)
        sdr_elt.set(u"{%s}role" % PBD_XMLNS, u"from")
        root_elt.append(sdr_elt)
    # Receiver address
    if acc_doc.receiver is not None:
        rcv_elt = build_address_xml_element(acc_doc.receiver ,xml_ns=PBD_XMLNS)
        rcv_elt.set(u"{%s}role" % PBD_XMLNS, u"to")
        root_elt.append(rcv_elt)
    # Remarks
    for rmk in [r for r in acc_doc.remarks if r.strip() != u""]:
        rmk_elt = etree.SubElement(root_elt, u"{%s}remark" % PBD_XMLNS)
        rmk_elt.text = rmk
    # Items list
    lst_elt = etree.SubElement(root_elt, u"{%s}items-list" % PBD_XMLNS)
    for acc_item in acc_doc.acc_items:
        itm_elt = etree.SubElement(lst_elt, u"{%s}item" % PBD_XMLNS)
        if acc_item.holdback_rate is not None:
            itm_elt.set(u"{%s}holdback-rate" % PBD_XMLNS,
                        unicode(acc_item.holdback_rate))
            if acc_item.holdback_on_vat:
                itm_elt.set(u"{%s}holdback-on-vat" % PBD_XMLNS, u"yes")
            else:
                itm_elt.set(u"{%s}holdback-on-vat" % PBD_XMLNS, u"no")
        qty_elt = etree.SubElement(itm_elt, u"{%s}qty" % PBD_XMLNS)
        qty_elt.set("{%s}digits" % PBD_XMLNS,
                    unicode(acc_item.quantity_digits))
        qty_elt.text = unicode(acc_item.quantity)
        desc_elt = etree.SubElement(itm_elt, u"{%s}description" % PBD_XMLNS)
        tit_elt = etree.SubElement(desc_elt, u"{%s}title" % PBD_XMLNS)
        tit_elt.text = acc_item.title
        for detail in [d for d in acc_item.details if d.strip() != u""]:
            dtl_elt = etree.SubElement(desc_elt, u"{%s}detail" % PBD_XMLNS)
            dtl_elt.text = detail
        upr_elt = etree.SubElement(itm_elt, u"{%s}unit-price" % PBD_XMLNS)
        upr_elt.set("{%s}digits" % PBD_XMLNS,
                    unicode(acc_item.unit_price_digits))
        upr_elt.text = unicode(acc_item.unit_price)
        if acc_item.vat_rate is not None:
            vtr_elt = etree.SubElement(itm_elt, u"{%s}vat-rate" % PBD_XMLNS)
            vtr_elt.text = unicode(acc_item.vat_rate)
    # Charged downpayment and issued debits (only for bills)
    if acc_doc.__class__ is Bill:
        for xml_elt_name, accdoc_list in \
                                      [(u"{%s}charged-downpayment" % PBD_XMLNS,
                                        acc_doc.charged_downpayments),
                                       (u"{%s}issued-debit" % PBD_XMLNS,
                                        acc_doc.issued_debits)]:
            for prev_doc in accdoc_list:
                pvd_elt = etree.SubElement(root_elt, xml_elt_name)
                pvd_elt.set(u"{%s}id" % PBD_XMLNS,
                            prev_doc.accdoc_id)
                pvd_elt.set(u"{%s}date" % PBD_XMLNS,
                            prev_doc.accdoc_date)
                pvd_elt.set(u"{%s}total" % PBD_XMLNS,
                            unicode(prev_doc.total))
                pvd_elt.set(u"{%s}vat" % PBD_XMLNS,
                            unicode(prev_doc.vat_amount))
    # Payment terms (only for bills, claim-forms, downpayments)
    if acc_doc.__class__ in [Bill, ClaimForm, Downpayment]:
        pmt_elt = etree.SubElement(root_elt, u"{%s}payment-terms" % PBD_XMLNS)
        pmt_elt.text = acc_doc.payment_terms
    # Finally returns the XML result
    return root_elt




# -*- coding: utf-8 -*-
"""
``pybill.lib.entities.accounting_docs.bill`` module contains the definition
of the entity representing a bill.

.. autoclass:: Bill
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib import ACCURACY
from pybill.lib.entities.accounting_docs.abstract import GenericAccountingDoc


class Bill(GenericAccountingDoc):
    """
    Entity class representing a bill.

    .. attribute:: charged_downpayments

       Attribute containing the list of the downpayments already charged in
       previous bills. The totals of these downpayments will be discarded from
       the total to be paid.

       type: list of :class:`~pybill.lib.entities.accounting_docs.utils.PreviousAccountingDoc`

    .. attribute:: issued_debits

       Attribute containing the list of the debits previously issued. The totals
       of these debits will be discarded from the total to be paid.

       type: list of :class:`~pybill.lib.entities.accounting_docs.utils.PreviousAccountingDoc`

    .. attribute:: payment_terms

       Attribute defining the description of the payment terms. This description
       will be displayed after the bank data that is read from the configuration
       file.
                         
       type: :class:`unicode`

    . automethod:: __init__
    """
    def __init__(self, bill_id):
        """
        Initializes an empty object.

        :parameter bill_id: Identifier of this accounting document.
        :type bill_id: :class:`unicode`
        """
        GenericAccountingDoc.__init__(self)
        self.charged_downpayments = []
        self.issued_debits = []
        self.payment_terms = u""
        # Filling the object with available data
        self.id = bill_id
        
    def get_downpayments_and_debits(self):
        """
        Computes the total of the already charged downpayments and the already
        issued debits.

        This total will be substracted at the end of the bill in order to
        compute the total to be paid.

        :returns: The total of downpayments and debits.
        :rtype: :class:`float`
        """
        total_dwp_dbt = 0.0
        for pad in (self.charged_downpayments + self.issued_debits):
            total_dwp_dbt += round(pad.total, ACCURACY)
        return round(total_dwp_dbt, ACCURACY)
            
    def get_to_be_paid_holdback_free(self):
        """
        Computes the total without holdback that must be paid.

        This total depends, of course, on the including taxes total but also on
        the holdbacks, the downpayments and the debits.
        
        :returns: The total without holdback of the bill that must be paid.
        :rtype: :class:`float`
        """
        total = self.get_including_taxes_total()
        total_dwp_dbt = self.get_downpayments_and_debits()
        total_hbk = self.get_holdback_amount() + self.get_holdback_vat_amount()
        if (total - total_hbk) >= total_dwp_dbt:
            return round(total - total_hbk - total_dwp_dbt, ACCURACY)
        else:
            return 0.0

    def get_to_be_paid_holdback(self):
        """
        Computes the total of holdbacks that must be paid.

        This total depends, of course, on the including taxes total but also on
        the holdbacks, the downpayments and the debits.
        
        :returns: The total of holdbacks of the bill that must be paid.
        :rtype: :class:`float`
        """
        total = self.get_including_taxes_total()
        total_dwp_dbt = self.get_downpayments_and_debits()
        total_hbk = self.get_holdback_amount() + self.get_holdback_vat_amount()
        if (total - total_hbk) >= total_dwp_dbt:
            return round(total_hbk, ACCURACY)
        else:
            return round(total_hbk - (total_dwp_dbt - (total - total_hbk)),
                         ACCURACY)

    def get_to_be_paid_vat(self):
        """
        Computes the total of VAT amounts that must be paid.

        This amount depends on the VAT amounts but also on the charged
        downpayments and the issued debits.

        :returns: The total of VAT amounts that must be paid.
        :rtype: :class:`float`
        """
        vat_amount = self.get_including_taxes_total() - self.get_price_total()
        prev_vats = [round(pad.vat_amount, ACCURACY) for pad
                     in (self.charged_downpayments + self.issued_debits)]
        vat_amount -= sum(prev_vats)
        return round(vat_amount, ACCURACY)



# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlreaders`` package contains the classes and functions used to
read the XML files describing the accounting documents and to build the
corresponding entity objects.

This package contains classes to read XML files either in the
`PyBill Document 0.X` old format or in the `PyBill Document 1.0` latest format.
It also defines the public
:class:`~pybill.lib.xmlreaders.accdoc.AccDocXMLReader` class that manages
the reading of an XML file whatever its PBD format is. This public class is the
only class that should be used from the outside of this package. It is imported
and available in the namespace of this package.
"""
__docformat__ = "restructuredtext en"

# Imports the classes and constants that must be used from the outside.
from pybill.lib.xmlreaders.accdoc import AccDocXMLReader




# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib import ACCURACY
from pybill.lib.entities.accounting_docs import AccItem


class AccItemTest(TestCase):
    def setUp(self):
        """
        Called before each test from this class
        """
        self.item = AccItem()
        
    def test_empty_object(self):
        """
        Tests an empty sold item contains nothing and has totals of 0.0
        """
        self.assertEqual(self.item.quantity, 0.0)
        self.assertEqual(self.item.quantity_digits, 0)
        self.assertEqual(self.item.unit_price, 0.0)
        self.assertEqual(self.item.unit_price_digits, ACCURACY)
        self.assertEqual(self.item.vat_rate, None)
        self.assertEqual(self.item.holdback_rate, None)
        self.assertEqual(self.item.holdback_on_vat, False)
        self.assertEqual(self.item.get_price(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.item.get_vat_amount(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.item.get_including_taxes_price(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.item.get_holdback_amount(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.item.get_holdback_vat_amount(),
                          round(0.0, ACCURACY))

    def test_get_price(self):
        """
        Tests the computation of the price (quantity * unit price)
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.assertEqual(self.item.get_price(),
                          round(65.0, ACCURACY))

    def test_get_price_with_quantity_digits(self):
        """
        Tests the computation of the price (quantity * unit price) with a
        different number of digits for the quantity.
        """
        self.item.quantity = 5.1234
        self.item.quantity_digits = 2
        self.item.unit_price = 100.0
        self.assertEqual(self.item.get_price(),
                          round(512.00, ACCURACY))

    def test_get_price_with_unit_price_digits(self):
        """
        Tests the computation of the price (quantity * unit price) with a
        different number of digits for the unit price.
        """
        self.item.quantity = 50.0
        self.item.unit_price = 100.00438
        self.item.unit_price_digits = 4
        self.assertEqual(self.item.get_price(),
                          round(5000.22, ACCURACY))

    def test_get_price_with_holdback(self):
        """
        Tests the computation of the price when holdback rate is defined
        (quantity * unit price)
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.holdback_rate = 5.0
        self.assertEqual(self.item.get_price(),
                          round(65.0, ACCURACY))

    def test_get_price_with_holdback_on_vat(self):
        """
        Tests the computation of the price when holdback rate on vat is defined
        (quantity * unit price)
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_price(),
                          round(65.0, ACCURACY))

    def test_get_vat_amount(self):
        """
        Tests the computation of the vat amount
        (quantity * unit price * vat rate)
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.assertEqual(self.item.get_vat_amount(),
                          round(12.74, ACCURACY))

    def test_get_vat_amount_with_no_vat(self):
        """
        Tests the computation of the vat amount when no vat rate is defined
        (quantity * unit price * vat rate)
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.assertEqual(self.item.get_vat_amount(),
                          round(0.0, ACCURACY))

    def test_get_vat_amount_with_holdback(self):
        """
        Tests the computation of the vat amount when holdback rate is defined 
        (quantity * unit price * vat rate)
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.0
        self.assertEqual(self.item.get_vat_amount(),
                          round(12.74, ACCURACY))

    def test_get_vat_amount_with_holdback_on_vat(self):
        """
        Tests the computation of the vat amount when holdback rate on vat is
        defined 
        (quantity * unit price * vat rate)
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_vat_amount(),
                          round(12.74, ACCURACY))

    def test_get_including_taxes_price(self):
        """
        Tests the computation of the including taxes price
        (quantity * unit price * (1.0 + vat rate) )
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.assertEqual(self.item.get_including_taxes_price(),
                          round(77.74, ACCURACY))

    def test_get_including_taxes_price_with_no_vat(self):
        """
        Tests the computation of the including taxes price when no vat rate is
        defined
        (quantity * unit price * (1.0 + vat rate) )
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.assertEqual(self.item.get_including_taxes_price(),
                          round(65.0, ACCURACY))

    def test_get_including_taxes_price_with_holdback(self):
        """
        Tests the computation of the including taxes price when holdback rate is
        defined
        (quantity * unit price * (1.0 + vat rate) )
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.0
        self.assertEqual(self.item.get_including_taxes_price(),
                          round(77.74, ACCURACY))

    def test_get_including_taxes_price_with_holdback_on_vat(self):
        """
        Tests the computation of the including taxes price when holdback rate
        on vat is defined
        (quantity * unit price * (1.0 + vat rate) )
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_including_taxes_price(),
                          round(77.74, ACCURACY))

    def test_get_holdback_with_no_holdback(self):
        """
        Tests the computation of the holdback when no holdback is defined
        (quantity * unit price * (holdback rate) )
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.assertEqual(self.item.get_holdback_amount(),
                          round(0.0, ACCURACY))

    def test_get_holdback_with_holdback(self):
        """
        Tests the computation of the holdback when holdback is defined
        (quantity * unit price * (holdback rate) )
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.0
        self.assertEqual(self.item.get_holdback_amount(),
                          round(3.25, ACCURACY))

    def test_get_holdback_with_holdback_on_vat(self):
        """
        Tests the computation of the holdback when holdback on vat is defined
        (quantity * unit price * (holdback rate) )
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_holdback_amount(),
                          round(3.25, ACCURACY))

    def test_get_holdback_vat_with_no_holdback(self):
        """
        Tests the computation of the vat holdback when no holdback is defined
        (quantity * unit price * vat_rate * (holdback rate) )
        if holdback_on_vat is True
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_holdback_vat_amount(),
                          round(0.0, ACCURACY))

    def test_get_holdback_vat_with_holdback(self):
        """
        Tests the computation of the vat holdback when holdback is defined
        (quantity * unit price * vat_rate * (holdback rate) )
        if holdback_on_vat is False
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.0
        self.assertEqual(self.item.get_holdback_vat_amount(),
                          round(0.0, ACCURACY))

    def test_get_holdback_vat_with_holdback_on_vat(self):
        """
        Tests the computation of the vat holdback when holdback on vat is
        defined
        (quantity * unit price * vat_rate * (holdback rate) )
        if holdback_on_vat is True
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_holdback_vat_amount(),
                          round(0.637, ACCURACY))

    def test_get_holdback_vat_with_holdback_on_vat_and_no_vat(self):
        """
        Tests the computation of the vat holdback when holdback on vat is
        defined but no vat rate id defined
        (quantity * unit price * vat_rate * (holdback rate) )
        if holdback_on_vat is True
        """
        self.item.quantity = 5.0
        self.item.unit_price = 13.0
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_holdback_vat_amount(),
                          round(0.0, ACCURACY))

    def test_get_price_when_rounding(self):
        """
        Tests the computation of the price when rounding to ``ACCURACY`` is
        done.
        (quantity * unit price)
        """
        self.item.quantity = 5.00115
        self.item.unit_price = 13.0
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_price(),
                          round(65.0, ACCURACY))

    def test_get_vat_amount_when_rounding(self):
        """
        Tests the computation of the vat amount when rounding to ``ACCURACY``
        is done.
        (quantity * unit price * vat rate)
        """
        self.item.quantity = 5.00115
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6001
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_vat_amount(),
                          round(12.74, ACCURACY))

    def test_get_including_taxes_price_when_rounding(self):
        """
        Tests the computation of the including taxes price when rounding to
        ``ACCURACY`` is done.
        (quantity * unit price * (1.0 + vat rate) )
        """
        self.item.quantity = 5.00115
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6001
        self.item.holdback_rate = 5.0
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_including_taxes_price(),
                          round(77.74, ACCURACY))

    def test_get_holdback_when_rounding(self):
        """
        Tests the computation of the holdback when rounding to ``ACCURACY`` is
        done.
        (quantity * unit price * (holdback rate) )
        """
        self.item.quantity = 5.00115
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6
        self.item.holdback_rate = 5.004
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_holdback_amount(),
                          round(3.25, ACCURACY))

    def test_get_holdback_vat_when_rounding(self):
        """
        Tests the computation of the vat holdback when rounding to ``ACCURACY``
        is done.
        (quantity * unit price * vat_rate * (holdback rate) )
        if holdback_on_vat is True
        """
        self.item.quantity = 5.00115
        self.item.unit_price = 13.0
        self.item.vat_rate = 19.6001
        self.item.holdback_rate = 5.004
        self.item.holdback_on_vat = True
        self.assertEqual(self.item.get_holdback_vat_amount(),
                          round(0.637, ACCURACY))


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.flowgenerators.proforma_generator`` module defines the
flow generator for the pro-forma entities.

The flow generator generates the title, various paragraphes, the main table
detailling the items, and the total table. This flow content will be displayed
on various pages by the
:class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
document template with the platypus machinery.

.. autoclass:: ProFormaFlowGenerator
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib.pdfwriters.flowgenerators.abstract import AbstractFlowGenerator

from reportlab.lib.units import cm

from reportlab.platypus import Paragraph, Table, TableStyle

from pybill.lib.pdfwriters import style_little_center
from pybill.lib.pdfwriters.utils import escape_text


class ProFormaFlowGenerator(AbstractFlowGenerator):
    """
    Class generating the PDF flow content for a
    :class:`~pybill.lib.entities.accounting_docs.proforma.ProForma`.

    .. automethod:: _get_acc_doc_title

    .. automethod:: _get_bottom_left_element

    .. automethod:: _get_total_table_cell_contents
    """
    def _get_acc_doc_title(self, acc_doc):
        """
        Gives the title of the accounting document (here, a pro-forma).

        :parameter acc_doc: The current accounting document that must be
                            exported in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns: The accounting document title.
        :rtype: :class:`unicode`
        """
        return (u"<b>%s</b> %s %s"
             % (escape_text(acc_doc.cfg_data.get_local_phrase(u'pro-forma')),
                escape_text(acc_doc.cfg_data.get_local_phrase(u'valid-until')),
                escape_text(acc_doc.validity_date, True)))
    
    def _get_bottom_left_element(self, acc_doc):
        """
        Gives the element that must be displayed at the bottom left of
        the accounting document.

        In the case of a pro-forma, this element is a set of lines asking for
        the agreement of the receiver.
        
        :parameter acc_doc: The current accounting document that must be
                            exported in PDF (here, a pro-forma).
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns: A table with a single cell containing the agreement lines
                  read from the configuration. This table is to be displayed
                  in the bottom left of the PDF document.
        :rtype: :class:`reportlab.platypus.Flowable`
        """
        table_data = [[]]
        # First and single row contains a single cell with several paragraphs.
        cell = []
        for line in acc_doc.cfg_data.agreement_intro_lines:
            cell.append(Paragraph(escape_text(line), style_little_center))
        table_data[0].append(cell)
        # Builds the agreement table (with its style).
        agr_table = Table(table_data, [9*cm], None)
        agr_table.setStyle(
            TableStyle([ ('ALIGN',     (0,0), (-1,-1), 'CENTER'),
                         ('VALIGN',    (0,0), (-1,-1), 'TOP'),
                       ])
                           )
        return agr_table

    def _get_total_table_cell_contents(self, acc_doc):
        """
        Gives the contents of the cells of the total table.

        These contents will be used by the ``_get_total_table`` method to build
        the total table.

        :parameter acc_doc: The current accounting document that must be
                            exported in PDF (here a pro-forma).
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns:
            A list describing the cells of the total table. The first element is
            a list of unicode strings corresponding to the various lines of the
            name of the row; the second element is a float corresponding to the
            number written in the row; the third element is a specification of
            the style of the row (``u"bold"``, ``u"italic"``,
            ``u"bold italic"``, ``u""``). 
        :rtype: list of triples (list of :class:`unicode`, :class:`float`,
                :class:`unicode`)
        """
        contents = []
        contents.extend(
            self._build_cell_contents_for_totals(acc_doc))
        return contents



# -*- coding: utf-8 -*-
"""
``pybill.commandline.ui`` module define the User Interface for PyBill command
line.

This User Interface consists in a :func:`run` function that defines a parser for
the command line arguments and relies on a PyBill controller for running the
PyBill functions.

This module can be executed and just runs the :func:`run` function defined
below.

In this module, we define a unique PyBill controller that will be used by the
run function.

.. data:: controller

   Unique controller that is used by the :func:`run` function to do the actual
   PyBill work.

   type: :class:`~pybill.lib.controller.PyBillController`

This module also defines the run function that provides the User Interface for
the command line.

.. autofunction:: run
"""
__docformat__ = "restructuredtext en"

from optparse import OptionParser
from traceback import format_exc
from sys import stderr
import os.path as osp
from pybill.lib.controller import PyBillController
from pybill.lib.errors import PyBillException
from pybill.__pkginfo__ import version

"""
Controller that does the actual PyBill work.
"""
controller = PyBillController()


def run():
    """
    Function containing the User Interface for the command line.

    This function hardly relies on a
    :class:`~pybill.lib.controller.PyBillController` object for doing the
    actual work and only reads and passes the correct arguments.

    In order to manage the command line arguments, this function defines and
    uses an :class:`optparse.OptionParser` object.
    """
    # Defines the parser of the command line options
    usage = ("Usage: %prog [options] accounting_files")
    desc = ("Generate a PDF representation of accounting documents (bills, "
            "etc.) The accounting documents are XML documents in a 'PyBill "
            "Document' format. They can represent a bill, a claim form, a "
            "downpayment, a debit, or a pro-forma. PyBill converts them in PDF"
            " and can generate the corresponding accounting entries in "
            "'pycompta' XML format.")
    opt_parser = OptionParser(usage=usage, description=desc,
                              version="PyBill %s" % version)
    opt_parser.add_option("-c", "--config", action="store", type="string",
                          dest="config_file", default=None,
                          metavar="CFGFILE",
                          help=("Name of the config file to be used for "
                                "generating the PDFs of the accounting "
                                "documents. This config file overrides the "
                                "config specified in each accounting "
                                "document."))
    opt_parser.add_option("-k", "--convert", action="store_true",
                          dest="convert_accdocs", default=False,
                          help=("Convert the processed accounting documents "
                                "into the last 'PyBill Document' format and "
                                "save them."))
    opt_parser.add_option("-n", "--no-entries", action="store_false",
                          dest="gen_entries", default=True,
                          help=("Don't generate the accounting entries "
                                "corresponding to the processed accounting "
                                "documents."))
    opt_parser.add_option("-e", "--entries-file", action="store", type="string",
                          dest="entries_file", default="entries.xml",
                          metavar="XMLFILE",
                          help=("Name of the file where the generated "
                                "accounting entries will be saved in 'pycompta'"
                                " format."))
    opt_parser.add_option("--encoding", action="store", type="string",
                          dest="xml_encoding", default="UTF-8",
                          metavar="XML_ENCODING",
                          help=("Encoding used for the XML files outputted by "
                                "this program (e.g. entries file or converted "
                                "pybill document."))
    opt_parser.add_option("--debug", action="store_true",
                          dest="debug", default=False,
                          help=("Print debug information and don't catch "
                                "the exceptions."))
    # Parses the command line options
    (options, acc_files) = opt_parser.parse_args()
    # Actual run
    try:
        acc_docs = controller.build_entities_from_files(acc_files,
                                                        options.config_file)
        if options.convert_accdocs:
            controller.save_entities_to_files(
                [(doc, "%s%sxml" % (filename, osp.extsep))
                 for doc, filename in acc_docs], options.xml_encoding)
        controller.write_pdf_to_files(
            [(doc, "%s%spdf" % (filename, osp.extsep))
             for doc, filename in acc_docs])
        if options.gen_entries:
            controller.export_accounting_entries_to_xml_file(
                [doc for doc, _ in acc_docs], options.entries_file, 
                options.xml_encoding)
    except PyBillException, exc:
        if not options.debug:
            stderr.write("Error while running PyBill:\n")
            stderr.write("%s\n" % exc)
        else:
            raise
    except Exception, exc:
        if not options.debug:
            stderr.write("Unexpected error during execution!\n")
            stderr.write("%s" % format_exc(limit=1))
        else:
            raise


if __name__ == '__main__' :
    run()



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main
from lxml import etree

from pybill.lib.xmlreaders.accdoc_format_1_0 import AccDocFormat_1_0_Reader
from pybill.lib.entities.accounting_docs import ClaimForm
from pybill.lib.config.entities import ConfigData

from generic_test_formatreader import FormatReaderGenericTest

xml_data = """
<accounting-document type="claim-form"
                     format-version="PBD-1.0"
                     xmlns="http://www.logilab.org/2010/PyBillDocument">
  <metadata>
    <id>A1</id>
    <doc-ref>A2</doc-ref>
    <place>A3</place>
    <date num="2009-10-19">A4</date>
    <info name="A5">A6</info>
    <info name="A7">A8</info>
    <info name="A9">A10</info>
  </metadata>

<address role="from">
  <honorific>B1</honorific>
  <firstname>B2</firstname>
  <othername>B3</othername>
  <surname>B4</surname>
  <lineage>B5</lineage>
  <street>B6</street>
  <street>B7</street>
  <pob>B8</pob>
  <postcode>B9</postcode>
  <city>B10</city>
  <state>B11</state>
  <country>B12</country>
  <phone>B13</phone>
  <fax>B14</fax>
  <web>B15</web>
  <email>B16</email>
  <affiliation>
    <jobtitle>B17</jobtitle>
    <jobtitle>B18</jobtitle>
    <orgname>B19</orgname>
    <orgdiv>B20</orgdiv>
    <orgdiv>B21</orgdiv>
    <address>
      <street>B22</street>
      <street>B23</street>
      <pob>B24</pob>
      <postcode>B25</postcode>
      <city>B26</city>
      <state>B27</state>
      <country>B28</country>
      <phone>B29</phone>
      <fax>B30</fax>
      <web>B31</web>
      <email>B32</email>
    </address>
  </affiliation>
</address>

<address role="to">
  <honorific>C1</honorific>
  <firstname>C2</firstname>
  <othername>C3</othername>
  <surname>C4</surname>
  <lineage>C5</lineage>
  <street>C6</street>
  <street>C7</street>
  <pob>C8</pob>
  <postcode>C9</postcode>
  <city>C10</city>
  <state>C11</state>
  <country>C12</country>
  <phone>C13</phone>
  <fax>C14</fax>
  <web>C15</web>
  <email>C16</email>
  <affiliation>
    <jobtitle>C17</jobtitle>
    <jobtitle>C18</jobtitle>
    <orgname>C19</orgname>
    <orgdiv>C20</orgdiv>
    <orgdiv>C21</orgdiv>
    <address>
      <street>C22</street>
      <street>C23</street>
      <pob>C24</pob>
      <postcode>C25</postcode>
      <city>C26</city>
      <state>C27</state>
      <country>C28</country>
      <phone>C29</phone>
      <fax>C30</fax>
      <web>C31</web>
      <email>C32</email>
    </address>
  </affiliation>
</address>

<remark>D1</remark>
<remark>D2</remark>

<items-list>
 <item>
  <qty digits="5">501.0</qty>
  <description>
   <title>E2</title>
   <detail>E3</detail>
   <detail>E4</detail>
  </description>
  <unit-price>505.0</unit-price>
 </item>
 <item>
  <qty>601.0</qty>
  <description>
   <title>F2</title>
  </description>
  <unit-price digits="6">603.0</unit-price>
  <vat-rate>604.0</vat-rate>
 </item>
 <item holdback-rate="7.0">
  <qty>701.0</qty>
  <description>
   <title>G2</title>
  </description>
  <unit-price>703.0</unit-price>
  <vat-rate>704.0</vat-rate>
 </item>
 <item holdback-rate="8.0" holdback-on-vat="yes">
  <qty>801.0</qty>
  <description>
   <title>H2</title>
  </description>
  <unit-price>803.0</unit-price>
  <vat-rate>804.0</vat-rate>
 </item>
</items-list>

<payment-terms>I1</payment-terms>

</accounting-document>
"""


class AccDocFormat10ReaderClaimFormTest(TestCase, FormatReaderGenericTest):
    def setUp(self):
        """
        Called before each test from this class
        """
        self.cfg = ConfigData()
        self.reader = AccDocFormat_1_0_Reader()

    def test_load_data(self):
        """
        Tests the Format_1_0_Reader reads correctly an XML claim form document.
        """
        xml_root = etree.fromstring(xml_data)
        acc_doc = self.reader.read_accounting_doc(xml_root,
                                                  filename="X0.xml",
                                                  cfg=self.cfg)
        self.assertEqual(acc_doc.__class__, ClaimForm)
        self.assertEqual(acc_doc.payment_terms, "I1")
        self._check_doc_generic_content(acc_doc)


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main
from lxml import etree
import datetime
from pybill.lib.entities.addresses import PersonAddress, OrganisationAddress
from pybill.lib.errors import PyBillReadingException

from pybill.lib.xmlreaders.utils import (AccDocFormatAbstractReader,
                                         read_person_address,
                                         read_text_from_xml_elt)


class FakeAccDocFormatReader(AccDocFormatAbstractReader):
    """
    Class used for testing the methods of the ``AccDocFormatAbstractReader``
    abstract class.
    """
    pass


class UtilFunctionsTest(TestCase):

    def test_read_text_from_simple_xml_element(self):
        """
        Tests the reading of a text inside an XML element.
        """
        xml_data = """
<root>
  <elt>TEXT</elt>
</root>
"""
        xml_root = etree.fromstring(xml_data)
        text = read_text_from_xml_elt(xml_root, u"elt")
        self.assertEqual(text, u"TEXT")

    def test_read_text_from_multiple_xml_element(self):
        """
        Tests the reading of a text inside an XML element when there are
        several XML elements (should only read the text in the first element).
        """
        xml_data = """
<root>
  <elt>TEXT1</elt>
  <elt>TEXT2</elt>
  <elt>TEXT3</elt>
</root>
"""
        xml_root = etree.fromstring(xml_data)
        text = read_text_from_xml_elt(xml_root, u"elt")
        self.assertEqual(text, u"TEXT1")
        
    def test_read_text_from_empty_xml_element(self):
        """
        Tests the reading of a text inside an empty XML element.
        """
        xml_data = """
<root>
  <elt></elt>
</root>
"""
        xml_root = etree.fromstring(xml_data)
        text = read_text_from_xml_elt(xml_root, u"elt")
        self.assertEqual(text, u"")

    def test_read_text_from_empty_xml_element_with_default_value(self):
        """
        Tests the reading of a text inside an empty XML element when a default
        value is defined.
        """
        xml_data = """
<root>
  <elt></elt>
</root>
"""
        xml_root = etree.fromstring(xml_data)
        text = read_text_from_xml_elt(xml_root, u"elt", u"FALLBACK")
        self.assertEqual(text, u"")

    def test_read_text_from_absent_xml_element(self):
        """
        Tests the reading of a text inside an XML element that doesn't exist.
        """
        xml_data = """
<root>
  <other-elt>TEXT</other-elt>
</root>
"""
        xml_root = etree.fromstring(xml_data)
        text = read_text_from_xml_elt(xml_root, u"elt")
        self.assertEqual(text, u"")

    def test_read_text_from_absent_xml_element_with_default_value(self):
        """
        Tests the reading of a text inside an XML element that doesn't exist 
        when a default value is defined.
        """
        xml_data = """
<root>
  <other-elt>TEXT</other-elt>
</root>
"""
        xml_root = etree.fromstring(xml_data)
        text = read_text_from_xml_elt(xml_root, u"elt", u"FALLBACK")
        self.assertEqual(text, u"FALLBACK")

    def test_read_text_from_complex_xml_element(self):
        """
        Tests the reading of a text inside an XML element with namespaces and
        several levels.
        """
        xml_data = """
<root xmlns:ns0="http://my.namespace.org/">
  <elt>
    <ns0:elt>TEXT</ns0:elt>
  </elt>
</root>
"""
        xml_root = etree.fromstring(xml_data)
        text = read_text_from_xml_elt(xml_root, 
                                      u"elt/{http://my.namespace.org/}elt")
        self.assertEqual(text, u"TEXT")

    def test_complete_address_reading(self):
        """
        Tests the reading of an address that has all of its fields filled.
        """
        xml_data = """
<address>
  <honorific>A1</honorific>
  <firstname>A2</firstname>
  <othername>A3</othername>
  <surname>A4</surname>
  <lineage>A5</lineage>
  <street>A6</street>
  <street>A7</street>
  <pob>A8</pob>
  <postcode>A9</postcode>
  <city>A10</city>
  <state>A11</state>
  <country>A12</country>
  <phone>A13</phone>
  <fax>A14</fax>
  <web>A15</web>
  <email>A16</email>
  <affiliation>
    <jobtitle>A17</jobtitle>
    <jobtitle>A18</jobtitle>
    <orgname>A19</orgname>
    <orgdiv>A20</orgdiv>
    <orgdiv>A21</orgdiv>
    <address>
      <street>A22</street>
      <street>A23</street>
      <pob>A24</pob>
      <postcode>A25</postcode>
      <city>A26</city>
      <state>A27</state>
      <country>A28</country>
      <phone>A29</phone>
      <fax>A30</fax>
      <web>A31</web>
      <email>A32</email>
    </address>
  </affiliation>
</address>
        """
        xml_addr = etree.fromstring(xml_data)
        addr = read_person_address(xml_addr)
        self.assertEqual(addr.__class__, PersonAddress)
        self.assertEqual(addr.honorific, u"A1")
        self.assertEqual(addr.firstname, u"A2")
        self.assertEqual(addr.other_name, u"A3")
        self.assertEqual(addr.surname, u"A4")
        self.assertEqual(addr.lineage, u"A5")
        self.assertEqual(addr.streets, [u"A6", u"A7"])
        self.assertEqual(addr.post_box, u"A8")
        self.assertEqual(addr.post_code, u"A9")
        self.assertEqual(addr.city, u"A10")
        self.assertEqual(addr.state, u"A11")
        self.assertEqual(addr.country, u"A12")
        self.assertEqual(addr.phone, u"A13")
        self.assertEqual(addr.fax, u"A14")
        self.assertEqual(addr.web, u"A15")
        self.assertEqual(addr.email, u"A16")
        # affiliation
        self.assertEqual(addr.organisation.__class__, OrganisationAddress)
        self.assertEqual(addr.organisation.job_titles, [u"A17", u"A18"])
        self.assertEqual(addr.organisation.name, u"A19")
        self.assertEqual(addr.organisation.divisions, [u"A20", u"A21"])
        self.assertEqual(addr.organisation.streets, [u"A22", u"A23"])
        self.assertEqual(addr.organisation.post_box, u"A24")
        self.assertEqual(addr.organisation.post_code, u"A25")
        self.assertEqual(addr.organisation.city, u"A26")
        self.assertEqual(addr.organisation.state, u"A27")
        self.assertEqual(addr.organisation.country, u"A28")
        self.assertEqual(addr.organisation.phone, u"A29")
        self.assertEqual(addr.organisation.fax, u"A30")
        self.assertEqual(addr.organisation.web, u"A31")
        self.assertEqual(addr.organisation.email, u"A32")

    def test_no_affiliation_address_reading(self):
        """
        Tests the reading of an address with no affiliation and thus no
        organisation address embedded
        """
        xml_data = """
<address>
  <honorific>A1</honorific>
  <firstname>A2</firstname>
  <othername>A3</othername>
  <surname>A4</surname>
  <lineage>A5</lineage>
  <street>A6</street>
  <street>A7</street>
  <pob>A8</pob>
  <postcode>A9</postcode>
  <city>A10</city>
  <state>A11</state>
  <country>A12</country>
  <phone>A13</phone>
  <fax>A14</fax>
  <web>A15</web>
  <email>A16</email>
</address>
        """
        xml_addr = etree.fromstring(xml_data)
        addr = read_person_address(xml_addr)
        self.assertEqual(addr.__class__, PersonAddress)
        self.assertEqual(addr.honorific, u"A1")
        self.assertEqual(addr.firstname, u"A2")
        self.assertEqual(addr.other_name, u"A3")
        self.assertEqual(addr.surname, u"A4")
        self.assertEqual(addr.lineage, u"A5")
        self.assertEqual(addr.streets, [u"A6", u"A7"])
        self.assertEqual(addr.post_box, u"A8")
        self.assertEqual(addr.post_code, u"A9")
        self.assertEqual(addr.city, u"A10")
        self.assertEqual(addr.state, u"A11")
        self.assertEqual(addr.country, u"A12")
        self.assertEqual(addr.phone, u"A13")
        self.assertEqual(addr.fax, u"A14")
        self.assertEqual(addr.web, u"A15")
        self.assertEqual(addr.email, u"A16")
        # affiliation
        self.assertEqual(addr.organisation, None)

    def test_regular_address_reading(self):
        """
        Tests the reading of a common address (with few fields).
        """
        xml_data = """
<address>
  <honorific>A1</honorific>
  <firstname>A2</firstname>
  <othername/>
  <surname>A3</surname>
  <phone>A4</phone>
  <email>A5</email>
  <affiliation>
    <jobtitle>A6</jobtitle>
    <orgname>A7</orgname>
    <address>
      <street>A8</street>
      <postcode>A9</postcode>
      <city>A10</city>
    </address>
  </affiliation>
</address>
        """
        xml_addr = etree.fromstring(xml_data)
        addr = read_person_address(xml_addr)
        self.assertEqual(addr.__class__, PersonAddress)
        self.assertEqual(addr.honorific, u"A1")
        self.assertEqual(addr.firstname, u"A2")
        self.assertEqual(addr.other_name, u"")
        self.assertEqual(addr.surname, u"A3")
        self.assertEqual(addr.phone, u"A4")
        self.assertEqual(addr.email, u"A5")
        # affiliation
        self.assertEqual(addr.organisation.__class__, OrganisationAddress)
        self.assertEqual(addr.organisation.job_titles, [u"A6"])
        self.assertEqual(addr.organisation.name, u"A7")
        self.assertEqual(addr.organisation.streets, [u"A8"])
        self.assertEqual(addr.organisation.post_code, u"A9")
        self.assertEqual(addr.organisation.city, u"A10")

    def test_address_with_namespace_reading(self):
        """
        Tests the reading of an address that uses a namespace.
        """
        xml_data = """
<address xmlns="http://www.logilab.org/TEST">
  <honorific>A1</honorific>
  <firstname>A2</firstname>
  <othername>A3</othername>
  <surname>A4</surname>
  <lineage>A5</lineage>
  <street>A6</street>
  <street>A7</street>
  <pob>A8</pob>
  <postcode>A9</postcode>
  <city>A10</city>
  <state>A11</state>
  <country>A12</country>
  <phone>A13</phone>
  <fax>A14</fax>
  <web>A15</web>
  <email>A16</email>
  <affiliation>
    <jobtitle>A17</jobtitle>
    <jobtitle>A18</jobtitle>
    <orgname>A19</orgname>
    <orgdiv>A20</orgdiv>
    <orgdiv>A21</orgdiv>
    <address>
      <street>A22</street>
      <street>A23</street>
      <pob>A24</pob>
      <postcode>A25</postcode>
      <city>A26</city>
      <state>A27</state>
      <country>A28</country>
      <phone>A29</phone>
      <fax>A30</fax>
      <web>A31</web>
      <email>A32</email>
    </address>
  </affiliation>
</address>
        """
        xml_addr = etree.fromstring(xml_data)
        addr = read_person_address(xml_addr,
                                   xml_ns="http://www.logilab.org/TEST")
        self.assertEqual(addr.__class__, PersonAddress)
        self.assertEqual(addr.honorific, u"A1")
        self.assertEqual(addr.firstname, u"A2")
        self.assertEqual(addr.other_name, u"A3")
        self.assertEqual(addr.surname, u"A4")
        self.assertEqual(addr.lineage, u"A5")
        self.assertEqual(addr.streets, [u"A6", u"A7"])
        self.assertEqual(addr.post_box, u"A8")
        self.assertEqual(addr.post_code, u"A9")
        self.assertEqual(addr.city, u"A10")
        self.assertEqual(addr.state, u"A11")
        self.assertEqual(addr.country, u"A12")
        self.assertEqual(addr.phone, u"A13")
        self.assertEqual(addr.fax, u"A14")
        self.assertEqual(addr.web, u"A15")
        self.assertEqual(addr.email, u"A16")
        # affiliation
        self.assertEqual(addr.organisation.__class__, OrganisationAddress)
        self.assertEqual(addr.organisation.job_titles, [u"A17", u"A18"])
        self.assertEqual(addr.organisation.name, u"A19")
        self.assertEqual(addr.organisation.divisions, [u"A20", u"A21"])
        self.assertEqual(addr.organisation.streets, [u"A22", u"A23"])
        self.assertEqual(addr.organisation.post_box, u"A24")
        self.assertEqual(addr.organisation.post_code, u"A25")
        self.assertEqual(addr.organisation.city, u"A26")
        self.assertEqual(addr.organisation.state, u"A27")
        self.assertEqual(addr.organisation.country, u"A28")
        self.assertEqual(addr.organisation.phone, u"A29")
        self.assertEqual(addr.organisation.fax, u"A30")
        self.assertEqual(addr.organisation.web, u"A31")
        self.assertEqual(addr.organisation.email, u"A32")

    def test_address_with_manual_namespace_reading(self):
        """
        Tests the reading of an address (with few fields) that manually defines
        a namespace on some elements (only these elements should be read).
        """
        xml_data = """
<ns:address xmlns:ns="http://www.logilab.org/TEST">
  <ns:honorific>A1</ns:honorific>
  <firstname>A2</firstname>
  <ns:surname>A3</ns:surname>
  <ns:phone>A4</ns:phone>
  <ns:email>A5</ns:email>
  <ns:affiliation>
    <ns:jobtitle>A6</ns:jobtitle>
    <ns:orgname>A7</ns:orgname>
    <ns:address>
      <street>A8</street>
      <ns:postcode>A9</ns:postcode>
      <ns:city>A10</ns:city>
    </ns:address>
  </ns:affiliation>
</ns:address>
        """
        xml_addr = etree.fromstring(xml_data)
        addr = read_person_address(xml_addr,
                                   xml_ns="http://www.logilab.org/TEST")
        self.assertEqual(addr.__class__, PersonAddress)
        self.assertEqual(addr.honorific, u"A1")
        self.assertEqual(addr.firstname, u"")
        self.assertEqual(addr.surname, u"A3")
        self.assertEqual(addr.phone, u"A4")
        self.assertEqual(addr.email, u"A5")
        # affiliation
        self.assertEqual(addr.organisation.__class__, OrganisationAddress)
        self.assertEqual(addr.organisation.job_titles, [u"A6"])
        self.assertEqual(addr.organisation.name, u"A7")
        self.assertEqual(addr.organisation.streets, [])
        self.assertEqual(addr.organisation.post_code, u"A9")
        self.assertEqual(addr.organisation.city, u"A10")


class AccDocFormatReaderCommonTest(TestCase):
    def setUp(self):
        """
        Called before each test of this class.
        """
        self.reader = FakeAccDocFormatReader()
        self.reader.filename = u"fake.xml"
        
    def test_xpath_selection(self):
        """
        Tests the selection of XML nodes with XPath pointers.
        """
        xml_data = """
<root>
  <elt num="i">az</elt>
  <elt num="ii">12</elt>
  <elt num="iii">2010-05-13</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        slct = self.reader._select_from_xpaths(xml, [u"elt"])
        self.assertEqual(len(slct), 3)
        for i in range(3):
            self.assertEqual(slct[i], xml[i])

    def test_xpath_complex_selection(self):
        """
        Tests the selection of XML nodes with XPath pointers.
        """
        xml_data = """
<root>
  <elt num="i">az</elt>
  <elt num="ii">12</elt>
  <elt num="iii">2010-05-13</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        slct = self.reader._select_from_xpaths(xml, [u"elt/@*|elt/text()"])
        self.assertEqual(slct,
                          [u"i", u"az", u"ii", u"12", u"iii", u"2010-05-13"])

    def test_xpath_selection_with_several_xpaths(self):
        """
        Tests the selection of XML nodes with XPath pointers (the first one
        doesn't select anything).
        """
        xml_data = """
<root>
  <elt num="i">az</elt>
  <elt num="ii">12</elt>
  <elt num="iii">2010-05-13</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        slct = self.reader._select_from_xpaths(xml,
                                      [u"nothing", u"elt/text()", u"elt/@num"])
        self.assertEqual(slct, ["az", "12", "2010-05-13"])

    def test_xpath_selection_with_namespace(self):
        """
        Tests the selection of XML nodes with XPath pointers containing the
        PyBill XML namespace.
        """
        xml_data = """
<root xmlns:ns="http://www.logilab.org/2010/PyBillDocument">
  <ns:elt num="i">az</ns:elt>
  <elt num="ii">12</elt>
  <ns:elt num="iii">2010-05-13</ns:elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        slct = self.reader._select_from_xpaths(xml, [u"pbd:elt/text()"])
        self.assertEqual(slct, ["az", "2010-05-13"])

    def test_xpath_typical_selection_with_namespace(self):
        """
        Tests the selection of XML nodes as it will be typically used.
        """
        xml_data = """
<ns:root xmlns:ns="http://www.logilab.org/2010/PyBillDocument">
  <ns:elt num="i">az</ns:elt>
  <ns:elt num="ii">12</ns:elt>
  <ns:elt num="iii">2010-05-13</ns:elt>
</ns:root>
        """
        xml = etree.fromstring(xml_data)
        slct = self.reader._select_from_xpaths(xml, [u"pbd:elt/@pbd:num",
                                                     u"pbd:elt/@num"])
        self.assertEqual(slct, ["i", "ii", "iii"])

    def test_xpath_reading_of_float_number(self):
        """
        Tests the reading of a float value with an XPath pointer.
        """
        xml_data = """
<root>
  <elt>az</elt>
  <elt>012.700</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        val = self.reader._read_number_from_xpaths(xml, [u"elt[2]/text()"],
                                                   float)
        self.assertEqual(type(val), float)
        self.assertEqual(val, 12.7)

    def test_xpath_reading_of_integer_number(self):
        """
        Tests the reading of an integer value with an XPath pointer.
        """
        xml_data = """
<root>
  <elt>az</elt>
  <elt>0012</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        val = self.reader._read_number_from_xpaths(xml, [u"elt[2]/text()"],
                                                   int)
        self.assertEqual(type(val), int)
        self.assertEqual(val, 12)

    def test_xpath_reading_of_integer_converted_to_float(self):
        """
        Tests the reading of an integer as a float with an XPath pointer.
        """
        xml_data = """
<root>
  <elt>az</elt>
  <elt>0012</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        val = self.reader._read_number_from_xpaths(xml, [u"elt[2]/text()"],
                                                   float)
        self.assertEqual(type(val), float)
        self.assertEqual(val, 12.0)

    def test_xpath_reading_of_absent_number(self):
        """
        Tests the reading of an absent value with an XPath pointer.
        """
        xml_data = """
<root>
  <elt>az</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        val = self.reader._read_number_from_xpaths(xml, [u"elt[2]/text()"],
                                                   float)
        self.assertEqual(val, None)

    def test_use_default_value_when_xpath_reading_of_absent_number(self):
        """
        Tests the use of the default value when reading an absent value with
        an XPath pointer.
        """
        xml_data = """
<root>
  <elt>az</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        val = self.reader._read_number_from_xpaths(xml, [u"elt[2]/text()"],
                                                   float, 13.4)
        self.assertEqual(val, 13.4)

    def test_xpath_reading_when_impossible_conversion_to_float(self):
        """
        Tests the reading, with an XPath pointer, of a value that can't be
        converted to float (the reading should raise an exception).
        """
        xml_data = """
<root>
  <elt>az</elt>
  <elt>12,7</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        self.assertRaises(PyBillReadingException,
                          self.reader._read_number_from_xpaths,
                          xml, [u"elt[2]/text()"], float)

    def test_xpath_reading_when_impossible_conversion_to_int(self):
        """
        Tests the reading, with an XPath pointer, of a value that can't be
        converted to integer (the reading should raise an exception).
        """
        xml_data = """
<root>
  <elt>az</elt>
  <elt>12.0</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        self.assertRaises(PyBillReadingException,
                          self.reader._read_number_from_xpaths,
                          xml, [u"elt[2]/text()"], int)

    def test_xpath_reading_of_number_with_several_xpaths(self):
        """
        Tests the reading of a number with several XPath pointers (the first
        one doesn't point to anything).
        """
        xml_data = """
<root>
  <elt>az</elt>
  <elt>2</elt>
  <elt>12</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        val = self.reader._read_number_from_xpaths(xml,
                                                   [u"number/text()",
                                                    u"elt[3]/text()",
                                                    u"elt[2]/text()"],
                                                   float)
        self.assertEqual(type(val), float)
        self.assertEqual(val, 12.0)

    def test_xpath_reading_of_date(self):
        """
        Tests the reading, with an XPath pointer, of a date.
        """
        xml_data = """
<root>
  <elt>2009-10-20</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        date = self.reader._read_date_from_xpaths(xml, [u"elt/text()"])
        self.assertEqual(date, datetime.date(2009, 10, 20))

    def test_xpath_reading_of_normalized_date(self):
        """
        Tests the reading, with an XPath pointer, of a date that is in
        normalized format.
        """
        xml_data = """
<root>
  <elt>2009-10-20Z</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        date = self.reader._read_date_from_xpaths(xml, [u"elt/text()"])
        self.assertEqual(date, datetime.date(2009, 10, 20))

    def test_xpath_reading_of_local_date(self):
        """
        Tests the reading, with an XPath pointer, of a date that is locally
        expressed.
        """
        xml_data = """
<root>
  <elt>2009-10-20-05:00</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        date = self.reader._read_date_from_xpaths(xml, [u"elt/text()"])
        self.assertEqual(date, datetime.date(2009, 10, 20))

    def test_xpath_reading_of_date_when_wrong_format(self):
        """
        Tests the reading, with an XPath pointer, of a date that is wrongly
        written.
        """
        xml_data = """
<root>
  <elt>2009-10-2</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        self.assertRaises(PyBillReadingException,
                          self.reader._read_date_from_xpaths, xml,
                          [u"elt/text()"])

    def test_xpath_reading_of_date_when_bad_value(self):
        """
        Tests the reading, with an XPath pointer, of a date that contains a
        wrong value for the month.
        """
        xml_data = """
<root>
  <elt>2009-20-10</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        self.assertRaises(PyBillReadingException,
                          self.reader._read_date_from_xpaths, xml,
                          [u"elt/text()"])

    def test_xpath_reading_of_date_with_several_xpaths(self):
        """
        Tests the reading of a date with several XPath pointers (the first
        one doesn't point to anything).
        """
        xml_data = """
<root>
  <elt>2010-05-13</elt>
  <elt>2010-05-12</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        date = self.reader._read_date_from_xpaths(xml,
                                                  [u"date/text()",
                                                   u"elt[2]/text()",
                                                   u"elt[1]/text()"])
        self.assertEqual(date, datetime.date(2010, 05, 12))

    def test_xpath_reading_of_text(self):
        """
        Tests the reading of a text with an XPath pointer.
        """
        xml_data = """
<root>
  <elt>abc</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        text = self.reader._read_text_from_xpaths(xml, [u"elt/text()"])
        self.assertEqual(type(text), unicode)
        self.assertEqual(text, u"abc")

    def test_xpath_reading_of_text_with_several_xpaths(self):
        """
        Tests the reading of a text with several XPath pointers (the first
        one doesn't point to anything).
        """
        xml_data = """
<root>
  <elt>13</elt>
  <elt>1</elt>
</root>
        """
        xml = etree.fromstring(xml_data)
        text = self.reader._read_text_from_xpaths(xml,
                                                  [u"text/text()",
                                                   u"elt[2]/text()",
                                                   u"elt[1]/text()"])
        self.assertEqual(type(text), unicode)
        self.assertEqual(text, u"1")

        
# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()




# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlreaders.accdoc_format_1_0`` module defines the specialized
format reader used to build an accounting document from XML data in
`PyBill Document 1.0` format.

.. autoclass:: AccDocFormat_1_0_Reader
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib import ACCURACY, PBD_1_0, PBD_XMLNS
from pybill.lib.errors import PyBillReadingException

from pybill.lib.xmlreaders.utils import (AccDocFormatAbstractReader,
                                         read_person_address)

from pybill.lib.entities.accounting_docs import (Bill, ClaimForm, Downpayment,
                                                 ProForma, Debit)
from pybill.lib.entities.accounting_docs import AccItem, PreviousAccountingDoc


ACC_CLASSES = {u"bill": Bill,
               u"claim-form": ClaimForm,
               u"debit": Debit,
               u"downpayment": Downpayment,
               u"pro-forma": ProForma}


class AccDocFormat_1_0_Reader(AccDocFormatAbstractReader):
    """
    Class reading the XML data contained in an accounting document in
    `PyBill Document 1.0` format.

    This class instanciates and fills the accounting object corresponding to the
    document described in the XML data.

    .. automethod:: __init__
    """
    def __init__(self):
        """
        Initializes a new ``AccDocFormat_1_0_Reader`` object.
        """
        AccDocFormatAbstractReader.__init__(self)
    
    def read_accounting_doc(self, xml_root, filename="", cfg=None):
        """
        Reads the data contained in the XML and builds an accounting object.

        
        The class of the object (``Bill``, ``ClaimForm``, ``Debit``,
        ``Downpayment``, or ``ProForma``) depends on the data contained in the
        XML.
        
        :parameter xml_root: root element of the XML tree containing the
                             accounting document. The XML has already been
                             read with `lxml` parser.
        :type xml_root: :class:`lxml.etree.Element`

        :parameter filename: name of the file the XML root element was read
                             from. This name is used to display nicest errors.
        :type filename: :class:`unicode`

        :parameter cfg: Configuration object that might be useful for
                        format conversion. For reading format `PDB-1.0`, the
                        configuration object is not needed.
        :type cfg: :class:`~pybill.lib.config.entities.ConfigData`
        
        :returns: entity object containing the accounting document.
        :rtype: :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`
        """
        self.filename = filename
        accdoc_type = self._read_text_from_xpaths(xml_root,
                                                  [u"@pbd:type", u"@type"])
        if accdoc_type not in ACC_CLASSES:
            raise PyBillReadingException("Unknown type ('%s') of accounting "
                                         "document in %s file."
                                         %(accdoc_type, self.filename))
        if accdoc_type in [u"pro-forma"]:
            acc_doc = ACC_CLASSES[accdoc_type]()
        else:
            acc_doc = ACC_CLASSES[accdoc_type](
                  self._read_text_from_xpaths(xml_root,
                                              [u"pbd:metadata/pbd:id/text()"]))
        # Reading metadata
        acc_doc.origin_format = PBD_1_0
        acc_doc.doc_ref = self._read_text_from_xpaths(xml_root,
                                          [u"pbd:metadata/pbd:doc-ref/text()"])
        acc_doc.date = self._read_text_from_xpaths(xml_root,
                                             [u"pbd:metadata/pbd:date/text()"])
        acc_doc.date_num = self._read_date_from_xpaths(xml_root,
                                            [u"pbd:metadata/pbd:date/@pbd:num",
                                             u"pbd:metadata/pbd:date/@num"],
                                            u"document standard date number")
        acc_doc.place = self._read_text_from_xpaths(xml_root,
                                            [u"pbd:metadata/pbd:place/text()"])
        for xml_info in self._select_from_xpaths(xml_root,
                                                 [u"pbd:metadata/pbd:info"]):
            xml_kw = self._read_text_from_xpaths(xml_info,
                                                 [u"@pbd:name", u"@name"])
            if xml_kw != u"":
                acc_doc.other_infos[xml_kw] = \
                             self._read_text_from_xpaths(xml_info, [u"text()"])
        # Reading adresses
        xml_sender_addrs = self._select_from_xpaths(xml_root,
                                             [u"pbd:address[@pbd:role='from']",
                                              u"pbd:address[@role='from']"])
        if len(xml_sender_addrs) > 0:
            acc_doc.sender = read_person_address(xml_sender_addrs[0],
                                                 xml_ns=PBD_XMLNS)
        xml_receiver_addrs = self._select_from_xpaths(xml_root,
                                               [u"pbd:address[@pbd:role='to']",
                                                u"pbd:address[@role='to']"])
        if len(xml_receiver_addrs) > 0:
            acc_doc.receiver = read_person_address(xml_receiver_addrs[0],
                                                   xml_ns=PBD_XMLNS)
        else:
            raise PyBillReadingException("Can't find receiver of accounting "
                                         "document in %s file whereas this "
                                         "piece of data is mandatory." 
                                         % self.filename)

        # Reading remarks
        for rmk in self._select_from_xpaths(xml_root, [u"pbd:remark"]):
            rem = self._read_text_from_xpaths(rmk, [u"text()"])
            if rem :
                acc_doc.remarks.append(rem)
        # Reading items list
        for xml_item in self._select_from_xpaths(xml_root,
                                                 [u"pbd:items-list/pbd:item"]):
            item = AccItem()
            item.quantity = self._read_number_from_xpaths(xml_item,
                                  [u"pbd:qty/text()"], float, 0.0, u"quantity")
            item.quantity_digits = self._read_number_from_xpaths(xml_item,
                                  [u"pbd:qty/@pbd:digits", u"pbd:qty/@digits"],
                                  int, 0, u"quantity digits")
            item.title = self._read_text_from_xpaths(xml_item,
                                         [u"pbd:description/pbd:title/text()"])
            for dtl in self._select_from_xpaths(xml_item,
                                              [u"pbd:description/pbd:detail"]):
                detail = self._read_text_from_xpaths(dtl, [u"text()"])
                if detail:
                    item.details.append(detail)
            item.unit_price = self._read_number_from_xpaths(xml_item,
                         [u"pbd:unit-price/text()"], float, 0.0, u"unit price")
            item.unit_price_digits = self._read_number_from_xpaths(xml_item,
                    [u"pbd:unit-price/@pbd:digits", u"pbd:unit-price/@digits"],
                    int, ACCURACY, u"unit price digits")
            item.vat_rate = self._read_number_from_xpaths(xml_item,
                            [u"pbd:vat-rate/text()"], float, None, u"VAT rate")
            item.holdback_rate =  self._read_number_from_xpaths(xml_item,
                                    [u"@pbd:holdback-rate", u"@holdback-rate"],
                                    float, None, u"holdback rate")
            hbk_on_vat = self._read_text_from_xpaths(xml_item,
                                [u"@pbd:holdback-on-vat", u"@holdback-on-vat"])
            item.holdback_on_vat = (hbk_on_vat.lower() == u"yes")
            acc_doc.acc_items.append(item)
        # Reading payment terms (for all the types except pro-forma and debit)
        if accdoc_type in [u"bill", u"claim-form", u"downpayment"]:
            acc_doc.payment_terms = self._read_text_from_xpaths(xml_root,
                                                 [u"pbd:payment-terms/text()"])
        # Reading charged downpayment and issued debits (bill specific infos)
        if accdoc_type == u"bill":
            for xml_elt_name, accdoc_list in \
                   [(u"pbd:charged-downpayment", acc_doc.charged_downpayments),
                    (u"pbd:issued-debit", acc_doc.issued_debits)]:
                for xml_elt in self._select_from_xpaths(xml_root,
                                                        [xml_elt_name]):
                    prev_doc = PreviousAccountingDoc()
                    prev_doc.total = self._read_number_from_xpaths(xml_elt,
                                 [u"@pbd:total", u"@total"], float, 0.0,
                                 u"%s total" % xml_elt_name.replace("-", " "))
                    prev_doc.vat_amount = self._read_number_from_xpaths(xml_elt,
                             [u"@pbd:vat", u"@vat"], float, 0.0,
                             u"%s vat amount" % xml_elt_name.replace("-", " "))
                    prev_doc.accdoc_id = self._read_text_from_xpaths(xml_elt,
                                                          [u"@pbd:id", u"@id"])
                    prev_doc.accdoc_date = self._read_text_from_xpaths(xml_elt,
                                                      [u"@pbd:date", u"@date"])
                    accdoc_list.append(prev_doc)
        # Reading downpayment percent (downpayment specific info)
        elif accdoc_type == u"downpayment":
            acc_doc.percent = self._read_number_from_xpaths(xml_root,
                                           [u"pbd:downpayment-percent/text()"],
                                           float, 30.0, u"downpayment percent")
        # Reading validity date (pro-forma specific info)
        elif accdoc_type == u"pro-forma":
            acc_doc.validity_date = self._read_text_from_xpaths(xml_root,
                                                 [u"pbd:validity-date/text()"])
        # Returns filled accounting document
        return acc_doc



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib import ACCURACY
from pybill.lib.entities.accounting_docs import ProForma

from generic_test_accounting_docs import AbstractAccDocTest


class ProFormaTest(TestCase, AbstractAccDocTest):
    """
    Tests on the pro-forma objects.
    """
    def setUp(self):
        """
        Called before each test from this class.
        """
        self.acc_doc = ProForma()

    def build_standard_example(self, with_holdbacks=False):
        """
        Creates a pro-forma document: a basic accounting document with a
        validity date.

        :parameter with_holdbacks: When set to ``True``, holdbacks are
                                   inserted in the items.
        :type with_holdbacks: bool
        """
        AbstractAccDocTest.build_standard_example(self,
                                                 with_holdbacks=with_holdbacks)
        self.acc_doc.validity_date = u"30/06/2009"

    def test_get_total_without_holdback_to_be_paid(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(0.0, ACCURACY))

    def test_get_total_holdback_to_be_paid(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(0.0, ACCURACY))

    def test_get_total_to_be_paid(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(0.0, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_when_rounding(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(0.0, ACCURACY))

    def test_get_total_holdback_to_be_paid_when_rounding(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid_when_rounding(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(0.0, ACCURACY))

    def test_get_total_to_be_paid_when_rounding(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(0.0, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_with_holdback(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(0.0, ACCURACY))

    def test_get_total_holdback_to_be_paid_with_holdback(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid_with_holdback(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(0.0, ACCURACY))

    def test_get_total_to_be_paid_with_holdback(self):
        """
        In the specific case of a pro-forma, nothing must be paid by the client
        (it's not realy a bill). Thus, the ``get_total_*_to_be_paid()``
        methods always return 0.0.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(0.0, ACCURACY))


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main
from StringIO import StringIO

from pybill.lib import PBD_1_0, PBD_0_X
from pybill.lib.accounting import CLIENT, CLIENT_HOLDBACK, PRODUCT, VAT
from pybill.lib.xmlreaders import AccDocXMLReader
from pybill.lib.errors import PyBillReadingException
from pybill.lib.config.entities import ConfigData
from pybill.lib.config import ConfigRegister
from pybill.lib.entities.accounting_docs import Bill

xml_data_1_0 = u"""
<accounting-document type="bill"
                     format-version="PBD-1.0"
                     xmlns="http://www.logilab.org/2010/PyBillDocument">

%s

<metadata>
  <id>A1</id>
  <doc-ref>A2</doc-ref>
  <place>A3</place>
  <date>A4</date>
</metadata>

<address role="from">
  <firstname>B1</firstname>
  <surname>B2</surname>
  <phone>B3</phone>
  <email>B4</email>
</address>

<address role="to">
  <firstname>C1</firstname>
  <surname>C2</surname>
  <affiliation>
    <orgname>C3</orgname>
    <address>
      <street>C4</street>
      <postcode>C5</postcode>
      <city>C6</city>
    </address>
  </affiliation>
</address>

<items-list>
 <item>
  <qty>41.0</qty>
  <description>
   <title>D2</title>
   <detail>D3</detail>
  </description>
  <unit-price>44.0</unit-price>
  <vat-rate>45.0</vat-rate>
 </item>
 <item>
  <qty>51.0</qty>
  <description>
   <title>E2</title>
  </description>
  <unit-price>43.0</unit-price>
  <vat-rate>44.0</vat-rate>
 </item>
</items-list>

<payment-terms>F1</payment-terms>

</accounting-document>
"""


xml_data_1_0_manual_ns = u"""
<pbd:accounting-document type="bill"
                         format-version="PBD-1.0"
                         xmlns:pbd="http://www.logilab.org/2010/PyBillDocument">

%s

<pbd:metadata>
  <pbd:id>A1</pbd:id>
  <pbd:doc-ref>A2</pbd:doc-ref>
  <pbd:place>A3</pbd:place>
  <pbd:date>A4</pbd:date>
</pbd:metadata>

<pbd:address role="from">
  <pbd:firstname>B1</pbd:firstname>
  <pbd:surname>B2</pbd:surname>
  <pbd:phone>B3</pbd:phone>
  <pbd:email>B4</pbd:email>
</pbd:address>

<pbd:address role="to">
  <pbd:firstname>C1</pbd:firstname>
  <pbd:surname>C2</pbd:surname>
  <pbd:affiliation>
    <pbd:orgname>C3</pbd:orgname>
    <pbd:address>
      <pbd:street>C4</pbd:street>
      <pbd:postcode>C5</pbd:postcode>
      <pbd:city>C6</pbd:city>
    </pbd:address>
  </pbd:affiliation>
</pbd:address>

<pbd:items-list>
 <pbd:item>
  <pbd:qty digits="3">41.0</pbd:qty>
  <pbd:description>
   <pbd:title>D2</pbd:title>
   <pbd:detail>D3</pbd:detail>
  </pbd:description>
  <pbd:unit-price digits="1">44.0</pbd:unit-price>
  <pbd:vat-rate>45.0</pbd:vat-rate>
 </pbd:item>
 <pbd:item holdback-rate="15.0" holdback-on-vat="yes">
  <pbd:qty>51.0</pbd:qty>
  <pbd:description>
   <pbd:title>E2</pbd:title>
  </pbd:description>
  <pbd:unit-price>43.0</pbd:unit-price>
  <pbd:vat-rate>44.0</pbd:vat-rate>
 </pbd:item>
</pbd:items-list>

<pbd:payment-terms>F1</pbd:payment-terms>

</pbd:accounting-document>
"""

xml_data_0_X = u"""
<bill id="A1"
      our-ref="A2"
      your-ref="A3"
      purch-ref="A4"
      place="A5" date="A6">

%s

<address role="from">
  <firstname>B1</firstname>
  <surname>B2</surname>
  <phone>B3</phone>
  <email>B4</email>
</address>

<address role="to">
  <firstname>C1</firstname>
  <surname>C2</surname>
  <affiliation>
    <orgname>C3</orgname>
    <address>
      <street>C4</street>
      <postcode>C5</postcode>
      <city>C6</city>
    </address>
  </affiliation>
</address>

<selling>
 <item>
  <qty>41.0</qty>
  <description>
   <title>D2</title>
   <detail>D3</detail>
  </description>
  <unit-price>44.0</unit-price>
  <vat>45.0</vat>
 </item>
 <item>
  <qty>51.0</qty>
  <description>
   <title>E2</title>
  </description>
  <unit-price>43.0</unit-price>
  <vat>44.0</vat>
 </item>
</selling>

<payment-terms>F1</payment-terms>

</bill>
"""

xml_data_1_0_short = u"""
<accounting-document type="bill"
                     format-version="PBD-1.0"
                     xmlns="http://www.logilab.org/2010/PyBillDocument">

%s

<metadata>
  <id>A1</id>
  <doc-ref>A2</doc-ref>
  <place>A3</place>
  <date>A4</date>
</metadata>

<address role="to">
  <firstname>C1</firstname>
  <surname>C2</surname>
  <street>C4</street>
  <postcode>C5</postcode>
  <city>C6</city>
</address>

<items-list>
 <item>
  <qty>41.0</qty>
  <description>
   <title>D2</title>
   <detail>D3</detail>
  </description>
  <unit-price>44.0</unit-price>
  <vat-rate>45.0</vat-rate>
 </item>
</items-list>

<payment-terms>F1</payment-terms>

</accounting-document>
"""

xml_data_0_X_short = u"""
<bill id="A1"
      our-ref="A2"
      your-ref="A3"
      purch-ref="A4"
      place="A5" date="A6">

%s

<address role="to">
  <firstname>C1</firstname>
  <surname>C2</surname>
  <street>C4</street>
  <postcode>C5</postcode>
  <city>C6</city>
</address>

<selling>
 <item>
  <qty>41.0</qty>
  <description>
   <title>D2</title>
   <detail>D3</detail>
  </description>
  <unit-price>44.0</unit-price>
  <vat>45.0</vat>
 </item>
</selling>

<payment-terms>F1</payment-terms>

</bill>
"""

xml_data_1_0_error = u"""
<accounting-document type="bill"
                     format-version="PBD-1.0"
                     xmlns="http://www.logilab.org/2010/PyBillDocument">

%s

<metadata>
  <id>A1</id>
  <doc-ref>A2</doc-ref>
  <place>A3</place>
  <date>A4</date>
</metadata>

<address role="from">
  <firstname>C1</firstname>
  <surname>C2</surname>
</address>

<items-list>
 <item>
  <qty>41.0</qty>
  <description>
   <title>D2</title>
   <detail>D3</detail>
  </description>
  <unit-price>44.0</unit-price>
  <vat-rate>45.0</vat-rate>
 </item>
</items-list>

<payment-terms>F1</payment-terms>

</accounting-document>
"""

xml_data_0_X_error = u"""
<bill id="A1"
      our-ref="A2"
      your-ref="A3"
      purch-ref="A4"
      place="A5" date="A6">

%s

<address role="from">
  <firstname>C1</firstname>
  <surname>C2</surname>
</address>

<selling>
 <item>
  <qty>41.0</qty>
  <description>
   <title>D2</title>
   <detail>D3</detail>
  </description>
  <unit-price>44.0</unit-price>
  <vat>45.0</vat>
 </item>
</selling>

<payment-terms>F1</payment-terms>

</bill>
"""

class AccDocXMLReaderTest(TestCase):
    def setUp(self):
        """
        Called before each test from this class
        """
        cfg_reg = ConfigRegister()
        self.cfg1 = ConfigData()
        self.cfg1.local_phrases = {u"your-ref-kw"    : u"X1",
                                   u"purch-ref-kw"   : u"X2",
                                   u"supplier-ref-kw": u"X3",
                                  }
        cfg_reg.configs[u"default"] = self.cfg1
        self.cfg2 = ConfigData()
        self.cfg2.local_phrases = {u"your-ref-kw"    : u"Y1",
                                   u"purch-ref-kw"   : u"Y2",
                                   u"supplier-ref-kw": u"Y3",
                                  }
        cfg_reg.configs[u"my-config"] = self.cfg2
        cfg_reg.configs[u"my/config.xml"] = self.cfg2
        self.cfg3 = ConfigData()
        self.cfg3.local_phrases = {u"your-ref-kw"    : u"Z1",
                                   u"purch-ref-kw"   : u"Z2",
                                   u"supplier-ref-kw": u"Z3",
                                  }
        cfg_reg.configs[u"another-config"] = self.cfg3
        self.reader = AccDocXMLReader(cfg_reg)

    def test_load_faulty_xml(self):
        """
        Tests the AccDocXMLReader raises an exception when trying to read
        a syntaxically incorrect XML file.
        """
        faulty_data = """
<accounting-document type="bill" format-version="PBD-1.0">
  <metadata>
</accounting-document>
                      """
        stream = StringIO(faulty_data)
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        self.assertRaises(PyBillReadingException,
                          self.reader.load_data, stream)

    def test_read_doc_with_no_config(self):
        """
        Tests the AccDocXMLReader reads an accounting document that doesn't
        specify the configuration to be used (so default will be used).
        """
        stream = StringIO(xml_data_0_X % u"")
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.cfg_data, self.cfg1)
        self.assertEqual(acc_doc.origin_format, PBD_0_X)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.other_infos, {u"X1": u"A3", u"X2": u"A4"})
        self.assertEqual(len(acc_doc.acc_items), 2)
        self.assertEqual(acc_doc.acc_items[0].title, u"D2")
        self.assertEqual(acc_doc.acc_items[1].unit_price, 43.0)
        self.assertEqual(acc_doc.payment_terms, u"F1")

    def test_read_doc_with_config(self):
        """
        Tests the AccDocXMLReader reads an accounting document that specifies
        in a processing instruction the configuration to be used.
        """
        stream = StringIO(xml_data_0_X % u'<?pybill config="my-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.cfg_data, self.cfg2)
        self.assertEqual(acc_doc.origin_format, PBD_0_X)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.other_infos, {u"Y1": u"A3", u"Y2": u"A4"})
        self.assertEqual(len(acc_doc.acc_items), 2)
        self.assertEqual(acc_doc.acc_items[0].title, u"D2")
        self.assertEqual(acc_doc.acc_items[1].unit_price, 43.0)
        self.assertEqual(acc_doc.payment_terms, u"F1")

    def test_read_doc_with_imposed_config(self):
        """
        Tests the AccDocXMLReader reads an accounting document with a
        configuration imposed from outside. Checks the internal specification
        of the configuration is not used.
        """
        stream = StringIO(xml_data_0_X % u'<?pybill config="another-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream, config_ref=u"my/config.xml")
        self.assertEqual(acc_doc.cfg_data, self.cfg2)
        self.assertEqual(acc_doc.origin_format, PBD_0_X)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.other_infos, {u"Y1": u"A3", u"Y2": u"A4"})
        self.assertEqual(len(acc_doc.acc_items), 2)
        self.assertEqual(acc_doc.acc_items[0].title, u"D2")
        self.assertEqual(acc_doc.acc_items[1].unit_price, 43.0)
        self.assertEqual(acc_doc.payment_terms, u"F1")

    def test_read_doc_with_accounts_definition(self):
        """
        Tests the AccDocXMLReader reads an accounting document that specifies
        in a processing instruction the impacted accounting accounts.
        """
        stream = StringIO(xml_data_0_X
                          % (u'<?accounts client="CLI" client-holdback="HLB" '
                             u'product="PRD" vat="VAT"?>'))
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.account_numbers,
                          {CLIENT : u"CLI", CLIENT_HOLDBACK: u"HLB",
                           PRODUCT: u"PRD", VAT            : u"VAT"})
        self.assertEqual(acc_doc.origin_format, PBD_0_X)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.other_infos, {u"X1": u"A3", u"X2": u"A4"})
        self.assertEqual(len(acc_doc.acc_items), 2)
        self.assertEqual(acc_doc.acc_items[0].title, u"D2")
        self.assertEqual(acc_doc.acc_items[1].unit_price, 43.0)
        self.assertEqual(acc_doc.payment_terms, u"F1")

    def test_read_doc_with_accounts_definition_and_config(self):
        """
        Tests the AccDocXMLReader reads an accounting document that specifies
        in a processing instruction the impacted accounting accounts.
        """
        stream = StringIO(xml_data_0_X
                          % (u'<?pybill config="my-config"?>\n'
                             u'<?accounts client="CLI" product="PRD"?>'))
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.account_numbers,
                          {CLIENT: u"CLI", PRODUCT: u"PRD"})
        self.assertEqual(acc_doc.origin_format, PBD_0_X)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.other_infos, {u"Y1": u"A3", u"Y2": u"A4"})
        self.assertEqual(len(acc_doc.acc_items), 2)
        self.assertEqual(acc_doc.acc_items[0].title, u"D2")
        self.assertEqual(acc_doc.acc_items[1].unit_price, 43.0)
        self.assertEqual(acc_doc.payment_terms, u"F1")

    def test_read_format_1_0_doc(self):
        """
        Tests the AccDocXMLReader correctly reads an accounting document
        in format 1.0.
        """
        stream = StringIO(xml_data_1_0 % u'<?pybill config="another-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.cfg_data, self.cfg3)
        self.assertEqual(acc_doc.origin_format, PBD_1_0)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.other_infos, {})
        self.assertEqual(len(acc_doc.acc_items), 2)
        self.assertEqual(acc_doc.acc_items[0].title, u"D2")
        self.assertEqual(acc_doc.acc_items[1].unit_price, 43.0)
        self.assertEqual(acc_doc.payment_terms, u"F1")

    def test_read_format_1_0_doc_manual_namespace(self):
        """
        Tests the AccDocXMLReader correctly reads an accounting document
        in format 1.0 with the namespace manually defined on all the elements
        but none of the attributes.
        """
        stream = StringIO(xml_data_1_0_manual_ns
                          % u'<?pybill config="another-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.cfg_data, self.cfg3)
        self.assertEqual(acc_doc.origin_format, PBD_1_0)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.other_infos, {})
        self.assertEqual(len(acc_doc.acc_items), 2)
        self.assertEqual(acc_doc.acc_items[0].title, u"D2")
        self.assertEqual(acc_doc.acc_items[0].quantity_digits, 3)
        self.assertEqual(acc_doc.acc_items[0].unit_price_digits, 1)
        self.assertEqual(acc_doc.acc_items[1].unit_price, 43.0)
        self.assertEqual(acc_doc.acc_items[1].holdback_rate, 15.0)
        self.assertEqual(acc_doc.acc_items[1].holdback_on_vat, True)
        self.assertEqual(acc_doc.payment_terms, u"F1")

    def test_read_format_0_X_doc(self):
        """
        Tests the AccDocXMLReader correctly reads an accounting document
        in format 0.X.
        """
        stream = StringIO(xml_data_0_X % u'<?pybill config="another-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.cfg_data, self.cfg3)
        self.assertEqual(acc_doc.origin_format, PBD_0_X) 
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.other_infos, {u"Z1": u"A3", u"Z2": u"A4"})
        self.assertEqual(len(acc_doc.acc_items), 2)
        self.assertEqual(acc_doc.acc_items[0].title, u"D2")
        self.assertEqual(acc_doc.acc_items[1].unit_price, 43.0)
        self.assertEqual(acc_doc.payment_terms, u"F1")

    def test_read_unknown_format_doc(self):
        """
        Tests the AccDocXMLReader doesn't read an accounting document in an
        unknown format.
        """
        wrong_data = u"""<wrong><?pybill config="default"?></wrong>"""
        stream = StringIO(wrong_data)
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        self.assertRaises(PyBillReadingException,
                          self.reader.load_data, stream)

    def test_read_format_1_0_short_doc(self):
        """
        Tests the AccDocXMLReader correctly reads an accounting document
        in format 1.0 with no sender data.
        """
        stream = StringIO(xml_data_1_0_short 
                          % u'<?pybill config="another-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.cfg_data, self.cfg3)
        self.assertEqual(acc_doc.origin_format, PBD_1_0)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.sender, None)
        self.assertEqual(acc_doc.receiver.honorific, u"")
        self.assertEqual(acc_doc.receiver.firstname, u"C1")
        self.assertEqual(acc_doc.receiver.surname, u"C2")
        self.assertEqual(acc_doc.receiver.streets, [u"C4"])
        self.assertEqual(acc_doc.receiver.post_code, u"C5")
        self.assertEqual(acc_doc.receiver.city, u"C6")
        self.assertEqual(acc_doc.receiver.organisation, None)

    def test_read_format_0_X_short_doc(self):
        """
        Tests the AccDocXMLReader correctly reads an accounting document
        in format 0.X with no sender data.
        """
        stream = StringIO(xml_data_0_X_short 
                          % u'<?pybill config="another-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        acc_doc = self.reader.load_data(stream)
        self.assertEqual(acc_doc.cfg_data, self.cfg3)
        self.assertEqual(acc_doc.origin_format, PBD_0_X)
        self.assertEqual(acc_doc.__class__, Bill)
        self.assertEqual(acc_doc.id, u"A1")
        self.assertEqual(acc_doc.sender, None)
        self.assertEqual(acc_doc.receiver.honorific, u"")
        self.assertEqual(acc_doc.receiver.firstname, u"C1")
        self.assertEqual(acc_doc.receiver.surname, u"C2")
        self.assertEqual(acc_doc.receiver.streets, [u"C4"])
        self.assertEqual(acc_doc.receiver.post_code, u"C5")
        self.assertEqual(acc_doc.receiver.city, u"C6")
        self.assertEqual(acc_doc.receiver.organisation, None)

    def test_read_format_1_0_error_doc(self):
        """
        Tests the AccDocXMLReader correctly raises an exception when reading
        an accounting document in format 1.0 with no receiver data.
        """
        stream = StringIO(xml_data_1_0_error
                          % u'<?pybill config="another-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        self.assertRaises(PyBillReadingException, 
                          self.reader.load_data, stream)

    def test_read_format_0_X_error_doc(self):
        """
        Tests the AccDocXMLReader correctly raises an exception when reading
        an accounting document in format 0.X with no receiver data.
        """
        stream = StringIO(xml_data_0_X_error 
                          % u'<?pybill config="another-config"?>')
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        self.assertRaises(PyBillReadingException, 
                          self.reader.load_data, stream)


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlreaders.accdoc_format_0_X`` module defines the specialized
format reader used to build an accounting document from XML data in
`PyBill Document 0.X` format.

The `PyBill Document 0.X` format is the format of the oldest versions of PyBill.

.. autoclass:: AccDocFormat_0_X_Reader
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib import PBD_0_X
from pybill.lib.errors import PyBillReadingException

from pybill.lib.xmlreaders.utils import (AccDocFormatAbstractReader,
                                         read_person_address)

from pybill.lib.entities.accounting_docs import (Bill, ClaimForm, Downpayment,
                                                 ProForma, Debit)
from pybill.lib.entities.accounting_docs import AccItem, PreviousAccountingDoc


ACC_CLASSES = {u"bill": Bill,
               u"claim-form":ClaimForm,
               u"debit": Debit,
               u"downpayment": Downpayment,
               u"pro-forma": ProForma}


class AccDocFormat_0_X_Reader(AccDocFormatAbstractReader):
    """
    Class reading the XML data contained in an accounting document in
    `PyBill Document 0.X` format.

    This class instanciates and fills the accounting object corresponding to the
    document described in the XML data.

    .. automethod:: __init__
    """
    def __init__(self):
        """
        Initializes a new `AccDocFormat_0_X_Reader` object.
        """
        AccDocFormatAbstractReader.__init__(self)
    
    def read_accounting_doc(self, xml_root, filename="", cfg=None):
        """
        Reads the data contained in the XML and builds an accounting object.
        
        The class of the object (``Bill``, ``ClaimForm``, ``Debit``,
        ``Downpayment``, or ``ProForma``) depends on the data contained in the
        XML.
        
        :parameter xml_root: root element of the XML tree containing the
                             accounting document. The XML has already been
                             read with `lxml` parser.
        :type xml_root: :class:`lxml.etree.Element`

        :parameter filename: name of the file the XML root element was read
                             from. This name is used to display nicest errors.
        :type filename: :class:`unicode`

        :parameter cfg: Configuration object that is useful for converting the
                        data in 0.X format.
        :type cfg: :class:`~pybill.lib.config.entities.ConfigData`
        
        :returns: entity object containing the accounting document.
        :rtype: :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`
        """
        self.filename = filename
        # The type is the local name of the root element (without namespace)
        accdoc_type = unicode(xml_root.tag[(xml_root.tag.find('}'))+1:])
        if accdoc_type not in ACC_CLASSES:
            raise PyBillReadingException("Unknown type ('%s') of accounting "
                                         "document in %s file."
                                         %(accdoc_type, self.filename))
        if accdoc_type in [u"pro-forma"]:
            acc_doc = ACC_CLASSES[accdoc_type]()
        else:
            acc_doc = ACC_CLASSES[accdoc_type](
                                 self._read_text_from_xpaths(xml_root, [u"@id"])
                                               )
        # Reading metadata
        acc_doc.origin_format = PBD_0_X
        acc_doc.doc_ref = self._read_text_from_xpaths(xml_root, [u"@our-ref"])
        acc_doc.date = self._read_text_from_xpaths(xml_root, [u"@date"])
        acc_doc.place = self._read_text_from_xpaths(xml_root, [u"@place"])
        for old_kw in [u"your-ref", u"purch-ref", u"supplier-ref"]:
            val = self._select_from_xpaths(xml_root, ["@%s" % old_kw])
            if len(val) > 0:
                if cfg is None:
                    kw_name = old_kw
                else:
                    kw_name = cfg.get_local_phrase(u"%s-kw" % old_kw)
                acc_doc.other_infos[kw_name] = unicode(val[0])
        # Reading adresses
        xml_sender_addrs = self._select_from_xpaths(xml_root,
                                                   [u"address[@role='from']"])
        if len(xml_sender_addrs) > 0:
            acc_doc.sender = read_person_address(xml_sender_addrs[0])
        xml_receiver_addrs = self._select_from_xpaths(xml_root,
                                                      [u"address[@role='to']"])
        if len(xml_receiver_addrs) > 0:
            acc_doc.receiver = read_person_address(xml_receiver_addrs[0])
        else:
            raise PyBillReadingException("Can't find receiver of accounting "
                                         "document in %s file whereas this "
                                         "piece of data is mandatory." 
                                         % self.filename)
        # Reading remarks
        for rmk in self._select_from_xpaths(xml_root, [u"remark"]):
            rem = self._read_text_from_xpaths(rmk, [u"text()"])
            if rem:
                acc_doc.remarks.append(rem)
        # Reading items list
        for xml_item in self._select_from_xpaths(xml_root, [u"selling/item"]):
            item = AccItem()
            item.quantity = self._read_number_from_xpaths(xml_item,
                                      [u"qty/text()"], float, 0.0, u"quantity")
            item.quantity_digits = 1
            item.title = self._read_text_from_xpaths(xml_item,
                                                 [u"description/title/text()"])
            for dtl in self._select_from_xpaths(xml_item,
                                                [u"description/detail"]):
                detail = self._read_text_from_xpaths(dtl, [u"text()"])
                if detail:
                    item.details.append(detail)
            item.unit_price = self._read_number_from_xpaths(xml_item,
                             [u"unit-price/text()"], float, 0.0, u"unit price")
            if (accdoc_type == u"debit" and
                unicode(xml_root.get(u"with-vat")) == u"no"):
                item.vat_rate = None
            else:
                item.vat_rate = self._read_number_from_xpaths(xml_item,
                                     [u"vat/text()"], float, 19.6, u"VAT rate")
            item.holdback_rate =  None
            item.holdback_on_vat = False
            acc_doc.acc_items.append(item)
        # Reading payment terms (for all the types except pro-forma and debit)
        if accdoc_type in [u"bill", u"claim-form", u"downpayment"]:
            acc_doc.payment_terms = self._read_text_from_xpaths(xml_root,
                                                     [u"payment-terms/text()"])
        # Reading charged downpayment (bill specific info)
        if accdoc_type == u"bill":
            for xml_chgdwp in self._select_from_xpaths(xml_root,
                                                     [u"charged-downpayment"]):
                chgdwp = PreviousAccountingDoc()
                chgdwp.total = self._read_number_from_xpaths(xml_chgdwp,
                        [u"text()"], float, 0.0, u"charged downpayment amount")
                chgdwp.accdoc_id = self._read_text_from_xpaths(xml_chgdwp,
                                                               [u"@bill-id"])
                chgdwp.accdoc_date = self._read_text_from_xpaths(xml_chgdwp,
                                                               [u"@bill-date"])
                acc_doc.charged_downpayments.append(chgdwp)
        # Reading issued debits (bill specific info)
        if accdoc_type == u"bill":
            for xml_issdbt in self._select_from_xpaths(xml_root,
                                                       [u"issued-debit"]):
                issdbt = PreviousAccountingDoc()
                issdbt.total = self._read_number_from_xpaths(xml_issdbt,
                               [u"text()"], float, 0.0, u"issued debit amount")
                issdbt.accdoc_id = self._read_text_from_xpaths(xml_issdbt,
                                                               [u"@debit-id"])
                issdbt.accdoc_date = self._read_text_from_xpaths(xml_issdbt,
                                                              [u"@debit-date"])
                acc_doc.issued_debits.append(issdbt)
        # Reading downpayment percent (downpayment specific info)
        if accdoc_type == u"downpayment":
            acc_doc.percent = self._read_number_from_xpaths(xml_root,
                            [u"@percent"], float, 30.0, u"downpayment percent")
        # Reading validity date (pro-forma specific info)
        if accdoc_type == u"pro-forma":
            acc_doc.validity_date = self._read_text_from_xpaths(xml_root,
                                                           [u"@validity-date"])
        # Returns filled accounting document
        return acc_doc



# -*- coding: utf-8 -*-
"""
``pybill.lib.config.register`` module defines the class that manages the
various configuration objects used during a PyBill run.

This class is the only public class of the ``pybill.lib.config`` package and
should be the only one used outside this package.

.. autoclass:: ConfigRegister
   :members:
   :undoc-members:
   :show-inheritance: 
"""
__docformat__ = "restructuredtext en"

import os
from os import path as osp

from pybill.lib.config.entities import ConfigData
from pybill.lib.config.xmlreaders import load_config


class ConfigRegister:
    """
    Class containing the various configuration objects read by PyBill.

    A configuration object is an instance of
    :class:`~pybill.lib.config.entities.ConfigData` class.  This register is
    used to get the correct configuration specified by an accounting document in
    its processing instruction. The register is also responsible for finding the
    configuration files and storing them after indexing them with their file
    name.

    Basically, the configuration files can be found in three places:

    - specified on the command-line (they are then stored in any directory),

    - specified in the processing instruction inside an accounting document and
      stored in the user configuration directory (\ ``~/.config/pybill/``\ ),

    - specified in the processing instruction inside an accounting document and
      stored in the system configuration directory
      (\ ``/etc/pybill/config/``\ ).

    .. attribute:: configs

        Attribute containing a dictionary that stores the configuration objects
        read from the disk, indexed by their reference.

        type: dictionary of :class:`~pybill.lib.config.entities.ConfigData`
              indexed with :class:`unicode` keys

    .. data:: SYSTEM_CONFIG_DIR

       Class constant containing the name of the system directory that contains
       the PyBill config files. Is currently u"/etc/pybill/config/".

       type: :class:`unicode`

    .. attribute:: USER_CONFIG_DIR

       Class constant containing the name of the user directory that contains
       the PyBill config files. Is currently u"~/.config/pybill/".

       type: :class:`unicode`

    .. attribute:: DEFAULT_CONFIG_REF

       Class constant containing the reference of the default
       configuration. This reference is the name of the file that contains this
       configuration. Is currently u"default".

       type: :class:`unicode`

    .. automethod:: __init__
    """
    SYSTEM_CONFIG_DIR = u"/etc/pybill/config/"
    USER_CONFIG_DIR = u"~/.config/pybill/"
    DEFAULT_CONFIG_REF = u"default"
    def __init__(self):
        """
        Intializes a new object.

        Actually there is only one single object for PyBill that manages the
        configurations.
        """
        self.configs = {}
        # Loads default config
        default_file = self.find_config_file(self.__class__.DEFAULT_CONFIG_REF)
        if default_file is None:
            self.configs[self.__class__.DEFAULT_CONFIG_REF] = ConfigData()
        else:
            self.store_config(self.__class__.DEFAULT_CONFIG_REF, default_file)

    def find_config_file(self, config_ref):
        """
        Finds the file containing the configuration whose name is
        ``config_ref``.

        ``config_ref`` can directly be a file or can be a reference to a file
        located in ``~/.config/pybill/`` or in ``/etc/pybill/config/``. In that
        last case, the file name is ``{config_ref}.xml``.

        If the file can't be found, returns None.

        :parameter config_ref: Reference of the configuration whose file must
                               be found.
        :type config_ref: :class:`unicode`

        :returns: The absolute path of the file containing the configuration
                  whose name is ``config_ref``. Returns ``None`` if the file
                  can't be found.
        :rtype: :class:`unicode`
        """
        if config_ref is not None and osp.exists(config_ref):
            return osp.abspath(config_ref)
        cfg_filename = u"%s.xml" % config_ref
        user_dir = osp.expanduser(self.__class__.USER_CONFIG_DIR)
        if osp.isdir(user_dir) and cfg_filename in os.listdir(user_dir):
            return unicode(osp.join(user_dir, cfg_filename))
        if (osp.isdir(self.__class__.SYSTEM_CONFIG_DIR)
            and cfg_filename in os.listdir(self.__class__.SYSTEM_CONFIG_DIR)):
            return unicode(osp.join(self.__class__.SYSTEM_CONFIG_DIR,
                                    cfg_filename))
        return None

    def store_config(self, config_ref, config_filename):
        """
        Stores a new configuration associated to the reference ``config_ref``.
        
        This configuration is described in the file ``config_filename`` and
        therefore must be read and stored in a ``ConfigData`` object before
        being stored in this object.

        :parameter config_ref: reference of the configuration that will be
                               added in this object.
        :type config_ref: :class:`unicode`

        :parameter config_filename: Absolute path of the file containing the
                                    configuration that will be added in this
                                    object.
        :type config_filename: :class:`unicode`
        """
        config = load_config(config_filename)
        config.reference = config_ref
        self.configs[config_ref] = config
    
    def get_config(self, config_ref):
        """
        Gets the configuration whose reference is ``config_ref``.

        If this configuration has already been read and is stored inside this
        object, directly returns it. Else, tries to find the file containing the
        desired configuration, reads it, stores it for further uses and returns
        it.

        :parameter config_ref: Reference of the desired configuration.
        :type config_ref: :class:`unicode`

        :returns: Configuration object whose name is ``config_ref``.
        :rtype: :class:`~pybill.lib.config.entities.ConfigData`
        """
        if config_ref not in self.configs:
            config_filename = self.find_config_file(config_ref)
            if config_filename is None:
                return self.configs[u"default"]
            else:
                self.store_config(config_ref, config_filename)
        return self.configs[config_ref]
